from raypier.tracer import RayTraceModel
from raypier.lenses import PlanoConvexLens
from raypier.sources import ConfocalRaySource

lens = PlanoConvexLens(centre=(20,0,0),
                       direction=(1,0,0),
                       curvature=25.0,
                       diameter=25.4,
                       CT=6.0)

source = ConfocalRaySource(focus = (0,0,0),
                           direction = (1,0,0),
                           working_dist= 100.0,
                           theta=15.0)

model = RayTraceModel(optics=[lens],
                      sources=[source])

model.configure_traits()



#    Copyright 2009, Teraview Ltd., Bryan Cole
#
#    This file is part of Raypier.
#
#    Raypier is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from traits.api import HasTraits, Array, Float, Complex,\
            Property, List, Instance, on_trait_change, Range, Any,\
            Tuple, Event, cached_property, Set, Int, Trait, Bool
from traitsui.api import View, Item, ListEditor, VSplit,\
            RangeEditor, ScrubberEditor, HSplit, VGroup, Heading
from tvtk.api import tvtk
from tvtk.pyface.scene_model import SceneModel
from tvtk.pyface.scene_editor import SceneEditor
import numpy
from itertools import chain, islice, tee

#from raypier.tracer import Optic, VTKOptic, normaliseVector, RaySegment,\
#             Traceable, NumEditor, dotprod, transformPoints, transformNormals

from raypier.bases import Optic, Traceable
from raypier.core.cfaces import PolygonFace, ExtrudedPlanarFace
from raypier.core.ctracer import FaceList


def pairwise(itr):
    a,b = tee(itr)
    fst = next(b)
    return zip(a, chain(b,[fst]))

class Extrusion(Optic):
    """a general flat-faced optic formed by extrusion 
    of a 2D polygon"""
    abstract=True
    profile = Array(shape=(None,2), dtype=numpy.double, transient=True)
    
    z_height_1 = Float(0.0)
    z_height_2 = Float(20.0)
    
    trace_ends = Bool(True, desc="include the end-faces in tracing")
    
    data_source = Instance(tvtk.ProgrammableSource, (), transient=True)
    
    extrude = Instance(tvtk.LinearExtrusionFilter, (), 
                       {'capping': True, 
                        'extrusion_type':'vector',
                        'vector': (0.,0.,1.)},
                        transient=True)
    
    def _faces_default(self):
        fl = FaceList(owner=self)
        fl.faces = self.make_faces()
        return fl
    
    def make_faces(self):
        z1 = self.z_height_1
        z2 = self.z_height_2
        profile = self.profile
        m = self.material
        #a plane is extruded differently from a 2d profile
        if profile.shape == (2,2):
            sides = [ExtrudedPlanarFace(owner=self, z1=z1, z2=z2, x1=profile[0,0], y1=profile[0,1], 
                        x2=profile[1,0], y2=profile[1,1], material=m)]
        else:
            sides = [ExtrudedPlanarFace(owner=self, z1=z1, z2=z2, x1=x1, y1=y1, 
                        x2=x2, y2=y2, material=m) for ((x2,y2),(x1,y1)) 
                                            in pairwise(profile)]
        if self.trace_ends:
            base = PolygonFace(owner=self, z_plane=z1,
                        xy_points=profile, material=m)
            top = PolygonFace(owner=self, z_plane=z2, material=m,
                        xy_points=profile, invert_normal=True)
            sides.extend([base, top])
        return sides
    
    @on_trait_change("z_height_1, z_height_2")
    def config_pipeline(self):
        self.extrude.scale_factor = self.z_height_2 - self.z_height_1
        self.faces.faces = self.make_faces()
        self.update=True
        
    def _trace_ends_changed(self):
        self.faces.faces = self.make_faces()
        self.update=True
        
    def _profile_changed(self):
        self.data_source.modified()
        self.faces.faces = self.make_faces()
        self.update=True
        
    def _vtk_profile(self):
        return self.profile
    
    def _pipeline_default(self):
        self.config_profile()
        self.config_pipeline()
        source = self.data_source
        def execute():
            xy = self._vtk_profile()
            z = numpy.ones(xy.shape[0]) * self.z_height_1
            points = numpy.column_stack((xy,z))
            
            cells = [list(range(len(z))),]
            
            output = source.poly_data_output
            output.points = points
            output.polys = cells
        source.set_execute_method(execute)
        print("Made PIPELINE")
        
        extrude = self.extrude
        extrude.input_connection = source.output_port
        
        t = self.transform
        transf = tvtk.TransformFilter(input_connection=extrude.output_port, 
                                      transform=t)
        return transf
    
    def config_profile(self):
        """abstract method to set the profile data"""
        pass


class Prism(Extrusion):
    name = "prism"
    abstract = False
    height = Float(20.0) #distance from front face to apex
    width = Float(20.0) #width of front face
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('trace_ends'),
                       Item('n_inside'),
                       Item('z_height_1', label="Z top"),
                       Item('z_height_2', label="Z bottom"),
                       Item('height'),
                       Item('width'),
                       )
                       )

    @on_trait_change("height, width")
    def config_profile(self):
        h = self.height
        w = self.width/2
        self.profile = [(-w,0),
                        (w,0),
                        (0,h)]
        
        
class Rhomboid(Extrusion):
    name = "rhomboid"
    abstract = False
    height = Float(15.0) #distance between parallel faces
    width = Float(15.0) #width of parallel faces
    slant = Float(45.0) #angle of the oblique faces
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('trace_ends'),
                       Item('n_inside'),
                       Item('z_height_1'),
                       Item('z_height_2'),
                       Item('height'),
                       Item('width'),
                       Item('slant'),
                       )
                       )

    @on_trait_change("height, width, slant")
    def config_profile(self):
        h = self.height/2
        angle = self.slant*numpy.pi/180
        s = h / numpy.tan(angle)
        w = self.width/2
        points = [(-w-s,-h),
                  (-w+s,h),
                  (w+s,h),
                  (w-s,-h)]
        points.reverse()
        self.profile = points
        
        
class TruncatedRightanglePrism(Extrusion):
    name = "truncated right-angle prism"
    abstract = False
    depth = Float #distance from front face to apex
    width = Float #width of front face
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('trace_ends'),
                       Item('n_inside'),
                       Item('depth'),
                       Item('width'),
                       Item('z_height_1', label="Z top"),
                       Item('z_height_2', label="Z bottom"),
                       )
                       )

    @on_trait_change("depth, width")
    def config_profile(self):
        h = self.depth
        w = self.width/2.
        self.profile = [(w,0),
                        (w, h-w),
                        (0,h),
                        (-w, h-w),
                        (-w,0)]
        

class LDLF(Extrusion):
    name = "Linear Dialectric Light Funnle"
    abstract = False
    slat_width= Float #width of slats
    ap_width = Float #width of exit apperture
    slant = Float #angle of slanted sides
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('trace_ends'),
                       Item('n_inside'),
                       Item('length'),
                       Item('slat_width'),
                       Item('ap_width'),
                       Item('slant')
                       )
                       )

    @on_trait_change("slat_width, ap_width, slant")
    def config_profile(self):
        theta = self.slant*numpy.pi/180.
        l = self.slat_width
        h = l*numpy.sin(theta)
        dis = l*numpy.cos(theta)
        w = self.ap_width/2
        self.profile = [(-w,0),
                        (w,0),
                        (w+dis,h),
                        (-w-dis,h)]
                        
class Sheet(Extrusion):
    #just a wrapper for the extrudedplanarface
    name = "Extruded Sheet"
    abstract = False
    x1 = Float 
    y1 = Float
    x2 = Float
    y2 = Float
    trace_ends = Bool(False)
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('trace_ends'),
                       Item('n_inside'),
                       Item('x1'),
                       Item('x2'),
                       Item('y1'),
                       Item('y2')
                       )
                       )

    @on_trait_change("x1,x2,y1,y2")
    def config_profile(self):
        
        self.profile = [(self.x1,self.y1),
                        (self.x2,self.y2)]

if __name__=="__main__":
    from ray_tracer import BuildRaySet, BeamStop, RayTraceModel
    
    input_rays = BuildRaySet(origin = (-7.42,15,0),
                             direction = (0,-1,0),
                             radius=1.0,
                             count=20)
    
    rhomboid = Rhomboid(n_inside = 1.764+0.0j,
                        orientation=0.0,
                        elevation=0.0,
                        centre=(0,0,0),
                        rotation=0,
                        length=10.0,
                        height=7.07,
                        width=14.0,
                        slant=45.0)
    
    beamstop = BeamStop(width=10,height=10,
                        centre=(7,-10,0),
                        direction=(0,1,0))
    
    #print "rhomboid", rhomboid.polydata
    print("beamstop", beamstop.polydata)
    
    model = RayTraceModel(optics=[rhomboid, beamstop], rays=input_rays)
    model.configure_traits()



from raypier.mirrors import PECMirror, RectMirror
from raypier.core.cmaterials import OpaqueMaterial
from tvtk.api import tvtk


class BeamStop(PECMirror):
    name = "Beam Stop"
    thickness = 0.5
    
    def _material_default(self):
        return OpaqueMaterial()

class RectTarget(RectMirror):
    name = "Beam Stop"
    thickness = 0.5

    vtkproperty = tvtk.Property(opacity = 0.4,
                             color = (0,0,0))

    def _material_default(self):
        return OpaqueMaterial()
    



"""
Passing object between threads with PyQt is a PITA. 
WX was so much easier...

Lifted this from Pyface.ui.gui module as its not exposed in the Pyface API.
"""

from pyface.qt import QtCore, QtGui



class FutureCall(QtCore.QObject):
    """ This is a helper class that is similar to the wx FutureCall class. """

    # Keep a list of references so that they don't get garbage collected.
    _calls = []

    # Manage access to the list of instances.
    _calls_mutex = QtCore.QMutex()

    # A new Qt event type for _FutureCalls
    _pyface_event = QtCore.QEvent.Type(QtCore.QEvent.registerEventType())

    def __init__(self, ms, callable, *args, **kw):
        super(FutureCall, self).__init__()

        # Save the arguments.
        self._ms = ms
        self._callable = callable
        self._args = args
        self._kw = kw

        # Save the instance.
        self._calls_mutex.lock()
        self._calls.append(self)
        self._calls_mutex.unlock()

        # Move to the main GUI thread.
        self.moveToThread(QtGui.QApplication.instance().thread())

        # Post an event to be dispatched on the main GUI thread. Note that
        # we do not call QTimer.singleShot here, which would be simpler, because
        # that only works on QThreads. We want regular Python threads to work.
        event = QtCore.QEvent(self._pyface_event)
        QtGui.QApplication.postEvent(self, event)

    def event(self, event):
        """ QObject event handler.
        """
        if event.type() == self._pyface_event:
            # Invoke the callable (puts it at the end of the event queue)
            QtCore.QTimer.singleShot(self._ms, self._dispatch)
            return True

        return super(FutureCall, self).event(event)

    def _dispatch(self):
        """ Invoke the callable.
        """
        try:
            self._callable(*self._args, **self._kw)
        finally:
            self._finished()

    def _finished(self):
        """ Remove the call from the list, so it can be garbage collected.
        """
        self._calls_mutex.lock()
        try:
            self._calls.remove(self)
        finally:
            self._calls_mutex.unlock()
            


#!/usr/bin/env python

# import pyximport
# pyximport.install()

import faulthandler
faulthandler.enable()

import sys
# sys.path.append('..')
from raypier.core.cmaterials import FullDielectricMaterial, Convert_to_SP, \
    TransparentMaterial, WaveplateMaterial, CoatedDispersiveMaterial, \
    BaseDispersionCurve, SingleLayerCoatedMaterial
from raypier.core.ctracer import Ray, RayCollection, norm, dotprod, subvv, cross
from raypier.dispersion import NondispersiveCurve
import unittest
from math import sqrt
import random
import numpy


P = lambda z: z.real**2 + z.imag**2

def ray_power(ray):
    """incident power per unit area"""
    P1 = (ray.E1_amp.real**2 + ray.E1_amp.imag**2)*ray.refractive_index.real
    P2 = (ray.E2_amp.real**2 + ray.E2_amp.imag**2)*ray.refractive_index.real
    ### We don't need to handle incident aspect area here as the ray already compensate for this
    #aspect = abs(dotprod(norm(ray.direction), normal))
    return (P1+P2)


class TestTransparentMaterial(unittest.TestCase):
    def test_preserve_polariasation(self):
        m = TransparentMaterial
        ray_idx = 123
        ray_in = Ray(origin=(0.0,0.0,0.0),
                     direction=(0.0,0.0,1.0),
                     E_vector=(1.0,0.0,0.0),
                     E1_amp = (1.0+2.0j),
                     E2_amp = (3.0+4.0j))

        out = RayCollection(1)
        point = (0.0,0.0,0.0)
        normal = (0.0,0.0,-1.0)





class TestConvertToSP(unittest.TestCase):
    def test_cross(self):
        v1 = (1.0,0.0,0.0)
        v2 = (0.0,1.0,0.0)
        self.assertEqual(cross(cross(v1,v2),v1), v2)

    def test_project_E(self):
        ray_in = Ray(origin=(-1.0,-2.0,-3.0),
                     direction=(1.0,2.0,3.0),
                     E_vector=(1.0,-2.0,0.0),
                     E1_amp = (1.0+2.0j),
                     E2_amp = (3.0+4.0j))
        ray_in.project_E(1.0,-2.0,0.0)

        E_vector = ray_in.E_vector
        E1_amp = ray_in.E1_amp
        E2_amp = ray_in.E2_amp

        ray_in.project_E(0.3,.7,0)
        ray_in.project_E(-0.8,0.235,0)
        ray_in.project_E(1.0,-2.0,0.0)

        self.assertEqual(E_vector, ray_in.E_vector)
        self.assertAlmostEqual(E1_amp, ray_in.E1_amp)
        self.assertAlmostEqual(E2_amp, ray_in.E2_amp)

    def test_convert_to_sp(self):
        ray_in = Ray(origin=(-1.0,-2.0,-3.0),
                     direction=(1.0,2.0,3.0),
                     E_vector=(1.0,-2.0,0.0),
                     E1_amp = (1.0+2.0j),
                     E2_amp = (3.0+4.0j))
        ray_in.project_E(1.5,-2.0,0.0)

        E_vector = ray_in.E_vector
        E1_amp = ray_in.E1_amp
        E2_amp = ray_in.E2_amp

        normal = (1.0, -1.0, 0.0)
        ray_out = Convert_to_SP(ray_in, normal)

        ray_out.project_E(1.5,-2.0,0.0)

        self.assertEqual(E_vector, ray_out.E_vector)
        self.assertAlmostEqual(E1_amp, ray_out.E1_amp)
        self.assertAlmostEqual(E2_amp, ray_out.E2_amp)


    def test_conserve_power(self):
        ray_in = Ray(origin=(-1.0,-2.0,-3.0),
                     direction=(1.0,2.0,3.0),
                     E_vector=(1.0,-2.0,0.0),
                     E1_amp = (1.0+2.0j),
                     E2_amp = (3.0+4.0j))

        P_in = P(ray_in.E1_amp) + P(ray_in.E2_amp)

        normal = (0.2,0.3,-1.1)

        ray_out = Convert_to_SP(ray_in, normal)

        P_out = P(ray_out.E1_amp) + P(ray_out.E2_amp)

        self.assertAlmostEqual(P_in, P_out)

        ray_out.project_E(1.0,-2.0,0.0)

        self.assertAlmostEqual(ray_out.E1_amp, ray_in.E1_amp)

    def test_normal_incidence(self):
        ray_in = Ray(origin=(-1.0,-2.0,-3.0),
                     direction=(1.0,2.0,3.0),
                     E_vector=(1.0,-2.0,0.0),
                     E1_amp = (1.0+2.0j),
                     E2_amp = (3.0+4.0j))

        P_in = P(ray_in.E1_amp) + P(ray_in.E2_amp)

        normal = (1.0,2.0,3.0)

        ray_out = Convert_to_SP(ray_in, normal)

        P_out = P(ray_out.E1_amp) + P(ray_out.E2_amp)

        self.assertAlmostEqual(P_in, P_out)

        ray_in.project_E(1.0,-2.0,0.0)
        ray_out.project_E(1.0,-2.0,0.0)

        self.assertEqual(ray_out.E_vector, ray_in.E_vector)
        self.assertAlmostEqual(ray_out.E1_amp, ray_in.E1_amp)
        self.assertAlmostEqual(ray_out.E2_amp, ray_in.E2_amp)


class TestWaveplateMaterial(unittest.TestCase):
    def test_apply_zero_retardance(self):
        ray_in = Ray(origin=(-1.0,-2.0,-3.0),
                     direction=(1.0,2.0,3.0),
                     E_vector=(1.0,-2.0,0.3),
                     E1_amp = (1.0+1.0j),
                     E2_amp = (2.0+-3.0j),
                     refractive_index=1.5)

        m = WaveplateMaterial(retardance=0.0)

        ray_out = m.apply_retardance(ray_in)

        self.assertEqual(ray_in.E1_amp, ray_out.E1_amp)
        self.assertEqual(ray_in.E2_amp, ray_out.E2_amp)

    def test_apply_quarter_wave_retardance(self):
        ray_in = Ray(origin=(-1.0,-2.0,-3.0),
                     direction=(1.0,2.0,3.0),
                     E_vector=(1.0,-2.0,0.3),
                     E1_amp = (2.0+2.0j),
                     E2_amp = (2.0+2.0j),
                     refractive_index=1.5)

        m = WaveplateMaterial(retardance=0.25)

        ray_out = m.apply_retardance(ray_in)

        self.assertEqual(ray_in.ellipticity, 0.0)
        self.assertEqual(abs(ray_out.ellipticity), 1.0)


    def test_apply_half_wave_retardance(self):
        ray_in = Ray(origin=(-1.0,-2.0,-3.0),
                     direction=(0.0,0.0,3.0),
                     E_vector=(1.0,0.0,0.3),
                     E1_amp = (1.0+1.0j),
                     E2_amp = (-1.0-1.0j),
                     refractive_index=1.5)

        m = WaveplateMaterial(retardance=0.5)

        ray_out = m.apply_retardance(ray_in)

        self.assertAlmostEqual(0.0, ray_out.ellipticity)

        ray_out.project_E(1.0,1.0,0.0)
        ray_in.project_E(1.0,1.0,0.0)
        print(ray_in.E_vector, ray_in.E1_amp, ray_in.E2_amp)
        print(ray_out.E_vector, ray_out.E1_amp, ray_out.E2_amp)
        self.assertAlmostEqual(ray_in.E1_amp, ray_out.E2_amp)
        self.assertAlmostEqual(ray_in.E2_amp, ray_out.E1_amp)

        ray_in = Ray(origin=(-1.0,-2.0,-3.0),
                     direction=(1.0,2.0,3.0),
                     E_vector=(1.0,-2.0,0.3),
                     E1_amp = (1.0+1.0j),
                     E2_amp = (2.0+-3.0j),
                     refractive_index=1.5)

        m = WaveplateMaterial(retardance=0.5)

        ray_out = m.apply_retardance(ray_in)

        self.assertAlmostEqual(abs(ray_in.ellipticity),
                          abs(ray_out.ellipticity))
        #self.assertEquals(ray_in.E2_amp, ray_out.E2_amp)


class TestFullDielectricMaterial(unittest.TestCase):
    def setUp(self):
        pass

    def test_eval_child_ray(self):
        ray_in = Ray(origin=(-1.0,-2.0,-3.0),
                     direction=(1.0,2.0,3.0),
                     E_vector=(1.0,-2.0,0.0),
                     E1_amp = (1.0+1.0j),
                     E2_amp = (2.0+0.0j),
                     refractive_index=1.5)

        mat = FullDielectricMaterial(n_inside=1.0,
                                     n_outside=1.5,
                                     reflection_threshold=-0.01,
                                     transmission_threshold=-0.01)

        out = RayCollection(8)
        ray_idx = 123
        point = (0.0,0.0,0.0)
        normal = (0.0,0.0,-1.0)
        tangent = (0.0,-1.0,0.0)

        mat.eval_child_ray(ray_in,
                           ray_idx,
                           point,
                           normal,
                           tangent,
                           out)

        self.assertEqual(len(out),2)

        aspect_in = dotprod(norm(ray_in.direction), norm(normal))
        P_in = (P(ray_in.E1_amp) + P(ray_in.E2_amp))*(ray_in.refractive_index.real)

        print("D_in:", norm((1.0,2.0,3.0)), "P_in:", P_in*aspect_in)
        for item in out:
            E1 = item.E1_amp
            E2 = item.E2_amp
            aspect_out = dotprod(norm(item.direction), norm(normal))
            print((P(E1) + P(E2))*(item.refractive_index.real)*aspect_out, item.refractive_index)

    def test_normal_incidence(self):
        n_out = 1.3
        n_in = 2.0

        ray_in = Ray(origin=(0.0,0.0,-1.0),
                     direction=(0.0,0.0,3.0),
                     E_vector=(5.0,0.0,0.0),
                     E1_amp = (1.0+1.0j),
                     E2_amp = (2.0+0.0j),
                     refractive_index=n_out)

        mat = FullDielectricMaterial(n_inside=n_in,
                                     n_outside=n_out,
                                     reflection_threshold=-0.01,
                                     transmission_threshold=-0.01)

        out = RayCollection(8)
        ray_idx = 123
        point = (0.0,0.0,0.0)
        normal = (0.0,0.0,-1.0)
        tangent = (0.0,-1.0,0.0)

        P_in = ray_power(ray_in)

        mat.eval_child_ray(ray_in,
                           ray_idx,
                           point,
                           normal,
                           tangent,
                           out)

        self.assertEqual(len(out),2)

        #textbook reflection and transmission coefficients
        R = ((n_in - n_out)/(n_in + n_out))**2
        T = (n_in/n_out)*((2*n_out/(n_out+n_in))**2)

        self.assertAlmostEqual(1, T+R)

        refl_pow = ray_power(out[0])
        self.assertAlmostEqual(R, refl_pow/P_in)

        trans_pow = ray_power(out[1])
        self.assertAlmostEqual(T, trans_pow/P_in)

        self.assertAlmostEqual(P_in, refl_pow+trans_pow)

        print(ray_in.E_left, ray_in.E_right)

    def test_circular_polarisation(self):
        n_out = 1.3
        n_in = 2.0

        ray_in = Ray(origin=(0.0,0.0,-1.0),
                     direction=(0.0,0.0,3.0),
                     E_vector=(5.0,0.0,0.0),
                     E1_amp = (2.0+2.0j),
                     E2_amp = (2.0-2.0j),
                     refractive_index=n_out)

        mat = FullDielectricMaterial(n_inside=n_in,
                                     n_outside=n_out,
                                     reflection_threshold=-0.01,
                                     transmission_threshold=-0.01)

        out = RayCollection(8)
        ray_idx = 123
        point = (0.0,0.0,0.0)
        normal = (0.0,0.0,-1.0)
        tangent = (0.0,-1.0,0.0)

        P_in = ray_power(ray_in)

        mat.eval_child_ray(ray_in,
                           ray_idx,
                           point,
                           normal,
                           tangent,
                           out)

        self.assertEqual(len(out),2)

        #textbook reflection and transmission coefficients
        R = ((n_in - n_out)/(n_in + n_out))**2
        T = (n_in/n_out)*((2*n_out/(n_out+n_in))**2)

        self.assertAlmostEqual(1, T+R)

        refl_pow = ray_power(out[0])
        self.assertAlmostEqual(R, refl_pow/P_in)

        trans_pow = ray_power(out[1])
        self.assertAlmostEqual(T, trans_pow/P_in)

        self.assertAlmostEqual(P_in, refl_pow+trans_pow)

        refl = out[0]
        trans = out[1]

        pwr = lambda x: (x*x.conjugate()).real
        P_left_in = pwr(ray_in.E_left)*ray_in.refractive_index
        P_right_in = pwr(ray_in.E_right)*ray_in.refractive_index

        P_left_refl = pwr(refl.E_left)*refl.refractive_index
        P_right_refl = pwr(refl.E_right)*refl.refractive_index

        P_left_trans = pwr(trans.E_left)*trans.refractive_index
        P_right_trans = pwr(trans.E_right)

        self.assertEqual(P_right_in, 0)
        self.assertEqual(P_left_refl, 0)
        self.assertEqual(P_right_trans, 0)

        #check for conservation of total power
        self.assertEqual(P_left_in, P_left_trans+P_right_refl)


    def test_brewster_angle(self):
        n_out = 1.2
        n_in = 2.0

        theta_B = numpy.arctan2(n_in, n_out)
        print("Brewsters angle:", theta_B*180/numpy.pi)

        y = numpy.sin(theta_B)
        z = numpy.cos(theta_B)

        #angle inside dielectric (for transmitted ray)
        theta_inside = numpy.arcsin( n_out * numpy.sin(theta_B) / n_in )
        trans_direction = norm((0.0, numpy.sin(theta_inside), numpy.cos(theta_inside)))

        ###input ray is directed into object
        ray_in = Ray(origin=(0.0,-y,-z),
                     direction=(0.0,y,z),
                     E_vector=(0.0,1.0,0.0),
                     E1_amp = (1.0+1.0j),
                     E2_amp = (0.0+0.0j),
                     refractive_index=n_out)

        mat = FullDielectricMaterial(n_inside=n_in,
                                     n_outside=n_out,
                                     reflection_threshold=-0.01,
                                     transmission_threshold=-0.01)

        out = RayCollection(8)
        ray_idx = 123
        point = (0.0,0.0,0.0)
        normal = (0.0,0.0,-1.0)
        tangent = (0.0,-1.0,0.0)

        P_in = ray_power(ray_in)

        mat.eval_child_ray(ray_in,
                           ray_idx,
                           point,
                           normal,
                           tangent,
                           out)

        self.assertEqual(len(out),2)

        refl_pow = ray_power(out[0])
        trans_pow = ray_power(out[1])

        self.assertAlmostEqual(0.0, refl_pow)
        self.assertAlmostEqual(refl_pow, out[0].power)

        self.assertAlmostEqual(P_in, trans_pow)

        ###check relfected ray direction
        self.assertAlmostEqual(abs(dotprod(norm(subvv(ray_in.direction,
                                                       out[0].direction)),
                                            normal)),
                                1)

        ###check transmitted ray direction
        self.assertAlmostEqual(abs(dotprod(out[1].direction,
                                            trans_direction)), 1)

class TestSingleLayerCoatedMaterial(unittest.TestCase):
    def test_normal_incidence(self):
        n_out = 1.3
        n_in = 2.0

        ray_in = Ray(origin=(0.0,0.0,-1.0),
                     direction=(0.0,0.0,3.0),
                     E_vector=(5.0,0.0,0.0),
                     E1_amp = (1.0+1.0j),
                     E2_amp = (2.0+0.0j),
                     refractive_index=n_out)

        mat = SingleLayerCoatedMaterial(n_inside=n_in,
                                     n_outside=n_out,
                                     n_coating=n_out,
                                     coating_thickness=0.0,
                                     reflection_threshold=-0.01,
                                     transmission_threshold=-0.01)
        mat.wavelengths = numpy.array([1.0])

        out = RayCollection(8)
        ray_idx = 123
        point = (0.0,0.0,0.0)
        normal = (0.0,0.0,-1.0)
        tangent = (0.0,-1.0,0.0)

        P_in = ray_power(ray_in)

        mat.eval_child_ray(ray_in,
                           ray_idx,
                           point,
                           normal,
                           tangent,
                           out)

        self.assertEqual(len(out),2)

        #textbook reflection and transmission coefficients
        R = ((n_in - n_out)/(n_in + n_out))**2
        
        ### Remember that the optical power is proportional to (n*E)^2
        T = ((n_in/n_out)*(2*n_out/(n_out+n_in))**2)

        self.assertAlmostEqual(1, T+R)

        refl_pow = ray_power(out[0])
        self.assertAlmostEqual(R, refl_pow/P_in)
        
        self.assertEqual(out[0].refractive_index, n_out)
        self.assertEqual(out[1].refractive_index, n_in)

        trans_pow = ray_power(out[1])
        self.assertAlmostEqual(T, trans_pow/P_in)

        self.assertAlmostEqual(P_in, refl_pow+trans_pow)

        print(ray_in.E_left, ray_in.E_right)
    
    def test_brewster_angle(self):
        n_out = 1.2
        n_in = 2.0

        theta_B = numpy.arctan2(n_in, n_out)
        print("Brewsters angle:", theta_B*180/numpy.pi)

        y = numpy.sin(theta_B)
        z = numpy.cos(theta_B)

        #angle inside dielectric (for transmitted ray)
        theta_inside = numpy.arcsin( n_out * numpy.sin(theta_B) / n_in )
        print("Internal angle:", theta_inside*180/numpy.pi)
        trans_direction = norm((0.0, numpy.sin(theta_inside), numpy.cos(theta_inside)))

        ###input ray is directed into object
        ray_in = Ray(origin=(0.0,-y,-z),
                     direction=(0.0,y,z),
                     E_vector=(0.0,1.0,0.0),
                     E1_amp = (1.0+1.0j),
                     E2_amp = (0.0+0.0j),
                     refractive_index=n_out)

        mat = SingleLayerCoatedMaterial(thickness=0.1,
                                        n_coating=1.5,
                                        n_inside=n_in,
                                     n_outside=n_out,
                                     reflection_threshold=-0.01,
                                     transmission_threshold=-0.01)
        mat.wavelengths=numpy.array([0.78])

        out = RayCollection(8)
        ray_idx = 123
        point = (0.0,0.0,0.0)
        normal = (0.0,0.0,-1.0)
        tangent = (0.0,-1.0,0.0)

        P_in = ray_power(ray_in)

        mat.eval_child_ray(ray_in,
                           ray_idx,
                           point,
                           normal,
                           tangent,
                           out)

        self.assertEqual(len(out),2)

        refl_pow = ray_power(out[0])
        trans_pow = ray_power(out[1])

        #self.assertAlmostEquals(0.0, refl_pow)
        #self.assertAlmostEquals(refl_pow, out[0].power)

        #self.assertAlmostEquals(P_in, trans_pow)

        ###check relfected ray direction
        self.assertAlmostEqual(abs(dotprod(norm(subvv(ray_in.direction,
                                                       out[0].direction)),
                                            normal)),
                                1)

        dir_out = out[1].direction
        theta_trans = 180*numpy.arctan2(dir_out[2], dir_out[1])/numpy.pi
        print("Transmitted angle:", theta_trans)
        ###check transmitted ray direction
        self.assertAlmostEqual(abs(dotprod(out[1].direction,
                                            trans_direction)), 1)

class TestDispersionMaterial(unittest.TestCase):
    def test_instantiate(self):
        m = CoatedDispersiveMaterial()

    def test_snells_law(self):
        n_out = 1.2
        n_in = 2.0

        d_out = NondispersiveCurve(n_out)
        d_in = NondispersiveCurve(n_in)
        d_coating = NondispersiveCurve(1.4)

        theta_B = numpy.arctan2(n_in, n_out)
        print("Brewsters angle:", theta_B*180/numpy.pi)

        y = numpy.sin(theta_B)
        z = numpy.cos(theta_B)

        #angle inside dielectric (for transmitted ray)
        theta_inside = numpy.arcsin( n_out * numpy.sin(theta_B) / n_in )
        trans_direction = norm((0.0, numpy.sin(theta_inside), numpy.cos(theta_inside)))

        ###input ray is directed into object
        ray_in = Ray(origin=(0.0,-y,-z),
                     direction=(0.0,y,z),
                     E_vector=(0.0,1.0,0.0),
                     E1_amp = (1.0+1.0j),
                     E2_amp = (0.0+0.0j),
                     refractive_index=n_out)
        print(ray_in.wavelength_idx)

        mat = CoatedDispersiveMaterial(dispersion_inside=d_in,
                                       dispersion_outside=d_out,
                                       coating_thickness=0.1,
                                       dispersion_coating=d_coating,
                                     reflection_threshold=0.01,
                                     transmission_threshold=0.01)
        wavelens = numpy.array([0.78], 'd')
        mat.wavelengths=wavelens

        out = RayCollection(8)
        ray_idx = 123
        point = (0.0,0.0,0.0)
        normal = (0.0,0.0,-1.0)
        tangent = (0.0,-1.0,0.0)

        P_in = ray_power(ray_in)

        mat.eval_child_ray(ray_in,
                           ray_idx,
                           point,
                           normal,
                           tangent,
                           out)

        self.assertEqual(len(out),2)

        refl_pow = ray_power(out[0])
        trans_pow = ray_power(out[1])

        #self.assertAlmostEquals(0.0, refl_pow)
        #self.assertAlmostEquals(refl_pow, out[0].power)

        #self.assertAlmostEquals(P_in, trans_pow)

        ###check relfected ray direction
        self.assertAlmostEqual(abs(dotprod(norm(subvv(ray_in.direction,
                                                       out[0].direction)),
                                            normal)),
                                1)

        ###check transmitted ray direction
        self.assertAlmostEqual(abs(dotprod(out[1].direction,
                                            trans_direction)), 1)


class TestDispersionCurve(unittest.TestCase):
    def test_sellmeier_2(self):
        coefs = numpy.array([0, 1.03961212, 0.00600069867, 0.231792344, 0.0200179144, 1.01046945, 103.560653])
        formula_id=2
        curve = BaseDispersionCurve(formula_id, coefs, absorption=1.0)

        wavelens = numpy.array([0.5,1.0,1.5])
        print(( curve.evaluate_n(wavelens) ))

    def test_bad_formula(self):
        coefs = numpy.array([0, 1.03961212, 0.00600069867, 0.231792344, 0.0200179144, 1.01046945, 103.560653])
        formula_id=10
        self.assertRaises(ValueError, BaseDispersionCurve, formula_id, coefs)

if __name__ == "__main__":
    unittest.main()



#    Copyright 2009, Teraview Ltd., Bryan Cole
#
#    This file is part of Raypier.
#
#    Raypier is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from traits.api import HasTraits, Array, Float, Complex,\
            Property, List, Instance, on_trait_change, Range, Any,\
            Tuple, Event, cached_property, Set, Int, Trait, Bool
from traitsui.api import View, Item, ArrayEditor, VGroup
from raypier.bases import ComplexEditor, NumEditor
from tvtk.api import tvtk
from tvtk.pyface.scene_model import SceneModel
from tvtk.pyface.scene_editor import SceneEditor
import numpy
from itertools import chain, islice, tee


from raypier.bases import Optic, Traceable
from raypier.core.cfaces import PolygonFace, ExtrudedBezierFace, ExtrudedPlanarFace
from raypier.core.ctracer import FaceList

import numpy as np
from scipy.interpolate import splprep, splev

def b_spline_to_bezier_series(tck, per = False):
  """Convert a parametric b-spline into a sequence of Bezier curves of the same degree.
 
  Inputs:
    tck : (t,c,k) tuple of b-spline knots, coefficients, and degree returned by splprep.
    per : if tck was created as a periodic spline, per *must* be true, else per *must* be false.
 
  Output:
    A list of Bezier curves of degree k that is equivalent to the input spline.
    Each Bezier curve is an array of shape (k+1,d) where d is the dimension of the
    space; thus the curve includes the starting point, the k-1 internal control
    points, and the endpoint, where each point is of d dimensions.


    from http://mail.scipy.org/pipermail/scipy-dev/2007-February/006651.html
    (actually, I got it from http://old.nabble.com/bezier-curve-through-set-of-2D-points-td27158642.html
    
  """
  from scipy.interpolate.fitpack import insert
  from numpy import asarray, unique, split, sum, transpose
  t,c,k = tck
  t = asarray(t)
  try:
    c[0][0]
  except:
    # I can't figure out a simple way to convert nonparametric splines to
    # parametric splines. Oh well.
    raise TypeError("Only parametric b-splines are supported.")
  new_tck = tck
  if per:
    # ignore the leading and trailing k knots that exist to enforce periodicity
    knots_to_consider = unique(t[k:-k])
  else:
    # the first and last k+1 knots are identical in the non-periodic case, so
    # no need to consider them when increasing the knot multiplicities below
    knots_to_consider = unique(t[k+1:-k-1])
  # For each unique knot, bring it's multiplicity up to the next multiple of k+1
  # This removes all continuity constraints between each of the original knots,
  # creating a set of independent Bezier curves.
  desired_multiplicity = k+1
  for x in knots_to_consider:
    current_multiplicity = sum(t == x)
    remainder = current_multiplicity%desired_multiplicity
    if remainder != 0:
      # add enough knots to bring the current multiplicity up to the desired multiplicity
      number_to_insert = desired_multiplicity - remainder
      new_tck = insert(x, new_tck, number_to_insert, per)
  tt,cc,kk = new_tck
  # strip off the last k+1 knots, as they are redundant after knot insertion
  bezier_points = transpose(cc)[:-desired_multiplicity]
  if per:
    # again, ignore the leading and trailing k knots
    bezier_points = bezier_points[k:-k]
  # group the points into the desired bezier curves
  return split(bezier_points, len(bezier_points) / desired_multiplicity, axis = 0)

def polate_3bezier(ctrl_pts,t):
    #evaluate a 3rd order bezier spline at t[0,1] given an array of four control points
    p0, p1,p2,p3 = ctrl_pts
    return p0*(1-t)**3 + p1*3*t*(1-t)**2 + p2*3*(1-t)*t**2 + p3*t**3


class Extruded_bezier(Optic):
    """an extrusion of a set of third order bezier curves.  Correct calculation of control point 
    location is done outside of this object, either in subclassed optical elements or in user's
    code."""
    name = "Bezier Spline"
    control_points = Array(shape=(None,4,2), dtype=numpy.double)


    z_height_1 = Float(-30.0)   #must be smaller than z_height_2
    z_height_2 = Float(30.0)
    

    trace_ends = Bool(False, desc="include the end-faces in tracing")
    trace_top = Bool(False, desc="include the end-faces in tracing")
    invert_normal = Bool(False,desc="Invert the direction of the surface normal vector")    

    data_source = Instance(tvtk.ProgrammableSource, ())
    
    extrude = Instance(tvtk.LinearExtrusionFilter, (), 
                       {'capping': False, 
                        'extrusion_type':'vector',
                        'vector': (0.,0.,1.)})
    
    traits_view = View(VGroup(
                       Traceable.uigroup,  
                       Item('z_height_1', editor=NumEditor),
                       Item('z_height_2', editor=NumEditor),
                       Item('invert_normal'),
                       Item('trace_top'),
                       )
                    )

    def __init__(self,*args,**kwargs):
        super(Extruded_bezier,self).__init__(*args,**kwargs)
        fl = FaceList(owner=self)
        fl.faces = self.make_faces()
        self.faces = fl

    '''#This was not working.  wrote above init to bypass this.  Leaving this here for referance.
    #enothought employee could not figure it out.
    def _faces_default(self):
        fl = FaceList(owner=self)
        fl.faces = self.make_faces()
        return fl
    '''
    def make_faces(self):
        z1 = self.z_height_1
        z2 = self.z_height_2
        ctl_pts = self.control_points
        m = self.material
        trace_ends = self.trace_ends
        trace_top = self.trace_top

        
        curves=[]
        #seems normal needs to be inverted.  All the time?  should there be a conditional test here?
        print("invert normal: ",self.invert_normal)
        curves.append(ExtrudedBezierFace(owner=self, beziercurves = ctl_pts, z_height_1 = self.z_height_1, z_height_2 = self.z_height_2, material=m, invert_normal = self.invert_normal))
        print(curves[0].invert_normal)
                            
        if trace_ends:
            """not perfect, just a polygon of the original profile"""
            profile = self.get_real_profile()
            #print "profile: ",profile
            prof =  np.column_stack((profile[:][0],profile[:][1]))
            base = PolygonFace(owner=self, z_plane=z1,
                        xy_points=prof, material=m)
            top = PolygonFace(owner=self, z_plane=z2, material=m,
                        xy_points=prof, invert_normal=True)
            curves.extend([base, top])
            
        if trace_top:
            curves.append( ExtrudedPlanarFace(owner=self, z1=z1, z2=z2, x1=ctl_pts[0][0][0], y1=ctl_pts[0][0][1], 
                    x2=ctl_pts[-1][-1][0], y2=ctl_pts[-1][-1][1], material=m) )
        return curves


            
    
    @on_trait_change("z_height_1, z_height_2")
    def config_pipeline(self):
        self.extrude.scale_factor = self.z_height_2 - self.z_height_1
        #print "extrude factor: ",  self.extrude.scale_factor
        self.faces.faces = self.make_faces()
        self.update=True
    
    @on_trait_change("invert_normal,trace_ends, trace_top")
    def trace_changed(self):
        self.faces.faces = self.make_faces()
        self.update=True

    '''
    def _control_points_changed(self):
        self.data_source.modified()
        self.faces.faces = self.make_faces()
        self.update=True
    '''
    def get_real_profile(self):
        #return the evaluated profile of the spline
        ts = np.linspace(0,1,7)    #splprep is defined of 0,1 of the parameter.  100 points
        profile = np.array([])
        profile.shape = (0,2)   
        for curve in self.control_points:
            for t in ts:
                profile = np.append(profile, [polate_3bezier(curve,t)],0)
        #print "real profile",profile
        return profile

    def _pipeline_default(self):
        source = self.data_source
        def execute():
            xy = self.get_real_profile()
            #print "extrude1:",profile[:3]
            #xy = np.column_stack((profile[:][0],profile[:][1]))
            #print "extrude2: ",xy[:3]
            z = numpy.ones(xy.shape[0]) * self.z_height_1
            points = numpy.column_stack((xy,z))
            
            cells = [list(range(len(z))),]
            
            output = source.poly_data_output
            output.points = points
            output.lines = cells
            print("cells: ",type(output))
        source.set_execute_method(execute)
        
        self.extrude.scale_factor = self.z_height_2 - self.z_height_1  #mm, put here because it wasn;t being initialized
        if self.trace_ends:
            print("drew ends")
            self.extrude.capping = True
        extrude = self.extrude
        extrude.input = source.output
        print("extrude: ",extrude.input)
        t = self.transform
        transf = tvtk.TransformFilter(input_connection=extrude.output_port, 
                                      transform=t)
        return transf
            

class Extruded_interpolant(Extruded_bezier):
    """a generalized extruded suface of a bezier spline approximating a surface"""
    
    #The profile is best with way too many points in it, because the spline
    #will determine the fewest number of ray tracing faces needed, but
    #but the end caps and visualization are polygon approximations
    #which are not speed critical use the profile directly.
    #
    #I haven't experiemnted to see what is excessive
    profile = Array(shape=(2,None), dtype=numpy.double)
    tck, uout = [None,None]      #the output of splprep 
    smoothness = Float(.0005)  #for splprep. found this number by trial and error. there are algorithms to guess better
    
    def make_faces(self):
        z1 = self.z_height_1
        z2 = self.z_height_2
        profile = self.profile
        m = self.material
        
        #convert profile to a list of bezier curves defined by three 2D points (knots)
        tck, uout = splprep(profile, s=self.smoothness, k=3, per=False)
        self.tck, self.uout = [tck,uout]
        self.control_points = b_spline_to_bezier_series(tck)
        print("splprep used ",len(self.control_points), " faces to make this spline")
        if len(self.control_points) > 30:
            print("!!! thats alot of faces.  try adjusting smoothness.")
            #need imperfection statistics

        return super(Extruded_interpolant,Extruded_interpolant).make_faces(self)                                
        
    def _smoothness_changed(self):
        self.faces.faces = self.make_faces()
        self.update=True
'''        
    def _profile_changed(self):
        self.faces.faces = self.make_faces()
        self.update=True
'''

'''
if __name__=="__main__":
    from ray_tracer import BuildRaySet, BeamStop, RayTraceModel
    
    input_rays = BuildRaySet(origin = (-7.42,15,0),
                             direction = (0,-1,0),
                             radius=1.0,
                             count=20)
    
    rhomboid = Rhomboid(n_inside = 1.764+0.0j,
                        orientation=0.0,
                        elevation=0.0,
                        centre=(0,0,0),
                        rotation=0,
                        length=10.0,
                        height=7.07,
                        width=14.0,
                        slant=45.0)
    
    beamstop = BeamStop(width=10,height=10,
                        centre=(7,-10,0),
                        direction=(0,1,0))
    
    #print "rhomboid", rhomboid.polydata
    print "beamstop", beamstop.polydata
    
    model = RayTraceModel(optics=[rhomboid, beamstop], rays=input_rays)
    model.configure_traits()
'''




"""
Traited wrappers for cfaces objects
"""


from .core.cfaces import ShapedPlanarFace, ShapedSphericalFace, ConicRevolutionFace, AsphericFace as AsphericFace_,\
            ShapedFace, AxiconFace as AxiconFace_, CylindericalFace as CylindericalFace_, DistortionFace as DistortionFace_,\
            SaddleFace as SaddleFace_
            
from .distortions import BaseDistortion
            
from .shapes import BaseShape
from .editors import NumEditor

from traits.api import HasStrictTraits, Instance, Float, Bool, observe, Event, Property
from traitsui.api import View, VGroup, Item, VGrid


class BaseFace(HasStrictTraits):
    """
    Base class for all high-level Face objects (not to be confused with raypier.core.cfaces classes). 
    
    raypier.Face objects wrap a lower level raypier.core.cfaces class which is the actual object that
    particiates in the raytracing operations. 
    
    The higher level objects which derive from this class provide a Traited wrapper with visualisation
    and UI for the object.
    """
    
    #: Local Z-axis intercept for the face 
    z_height = Float(0.0)
    
    #: An event-trait which fires whenever a face-parameter changes.
    updated = Event()
    
    #: Indicates if this face should be included in the ray-tracing. If not, it's purely cosmetic
    trace = Bool(True)
    
    #: If mirror is True, this face is reflective (i.e. doesn't use material info to transmit rays
    mirror = Bool(False)
    
    #: I'm not 100% sure if we need this. This inverts the direction of the face-normal.
    invert = Bool(False)
    
    #: Override this in subclasses with the specific face type required.
    cface = Instance(ShapedFace)
        
    def _invert_changed(self, vnew):
        self.cface.invert_normals = int(vnew)
        self.updated=True
        
    @observe("trace, mirror")
    def _params_changed(self, evt):
        self.updated=True
    

class PlanarFace(BaseFace):
    """
    A planar face. You probably guessed this from the name.
    """
    cface = Instance(ShapedPlanarFace, (), )
    
    traits_view = View(VGroup(Item("z_height", editor=NumEditor, tooltip="surface height in mm")))
    
    def __repr__(self):
        return f"<Planar Face: z={self.z_height}>"
    
    def _z_height_changed(self, znew):
        self.cface.z_height = znew
        self.updated=True
        
        
class SphericalFace(PlanarFace):
    #: Radius of curvature for the spherical face, in mm.
    curvature = Float(100.0)
    
    cface = Instance(ShapedSphericalFace, ())
    
    traits_view = View(VGroup(
            Item("z_height", editor=NumEditor, tooltip="surface height at x=0,y=0 in mm"),
            Item("curvature", editor=NumEditor, tooltip="radius of curvature, in mm")
        ))
    
    def __repr__(self):
        return f"<Spherical Face: z={self.z_height}, curvature={self.curvature}>"
    
    def _curvature_changed(self, cnew):
        self.cface.curvature = cnew
        self.updated=True
        
        
class AxiconFace(PlanarFace):
    
    #: Sets the gradient of the axicon face in terms of dz/dr. 
    gradient = Float(0.1)
    
    cface = Instance(AxiconFace_, ())
    
    traits_view = View(VGroup(
            Item("z_height", editor=NumEditor, tooltip="surface height at x=0,y=0 in mm"),
            Item("gradient", editor=NumEditor, tooltip="gradient of the sides of the axicon, dz/dr.")
        ))
    
    def _cface_default(self):
        return AxiconFace_(z_height=self.z_height, gradient=self.gradient)
        
    def _gradient_changed(self, vnew):
        self.cface.gradient = vnew
        self.updated=True
        
        
class SaddleFace(PlanarFace):
    """
    Creates a saddle-shaped face equivalent to a n=2,m=2 Zernike distortion.
    
    This is mainly useful in creating a highly astigmatic beam, as a way of verifying the
    correct calculation and propagation of astigmatic Gaussian modes. 
    """
    
    #: Sets the second order curvature of a face, i.e. d2z/dxdy.
    curvature = Float(1.0)
    
    cface = Instance(SaddleFace_, ())
    
    traits_view = View(VGroup(
            Item("z_height", editor=NumEditor, tooltip="surface height at x=0,y=0 in mm"),
            Item("curvature", editor=NumEditor, tooltip="curvature of the saddle-point")
        ))
    
    def _cface_default(self):
        return SaddleFace_(z_height=self.z_height, curvature=self.curvature)
        
    def _curvature_changed(self, vnew):
        self.cface.curvature = vnew
        self.updated=True
        
        
class CylindericalFace(SphericalFace):
    """
    A cylinderical face.
    """
    cface = Instance(CylindericalFace_)
    
    def __repr__(self):
        return f"<Cylinderical Face: z={self.z_height}, curvature={self.curvature}>"
    
    def _cface_default(self):
        return CylindericalFace_(radius=self.curvature, z_height=self.z_height)
    
    def _curvature_changed(self, cnew):
        self.cface.radius = cnew
        self.updated=True
    
        
class ConicFace(SphericalFace):
    """
    Creates a conic surface of revolution.
    """
    #: Conic constant.
    conic_const = Float(0.0)
    
    cface = Instance(ConicRevolutionFace, ())
    
    traits_view = View(VGroup(
            Item("z_height", editor=NumEditor, tooltip="surface height at x=0,y=0 in mm"),
            Item("curvature", editor=NumEditor, tooltip="radius of curvature, in mm"),
            Item("conic_const", editor=NumEditor, tooltip="conic constant for the surface"),
            Item("invert", editor=NumEditor, tooltip="Invert the outward sense for this surface")
        ))
    
    def _conic_const_changed(self, knew):
        self.cface.conic_const = knew
        self.updated=True
    
    
class AsphericFace(ConicFace):
    """
    A General aspheric face, defined in terms of a sag function given by a conic surface of revolution
    plus a sequence of even-order polynomial coefficients in r.
    
    The quadratice term is absent, being redundant since the conic surface includes a quadratic component.
    """
    #: A**4 term
    A4 = Float(0.0)
    
    #: A**6 term
    A6 = Float(0.0)
    
    #: A**8 term
    A8 = Float(0.0)
    
    #: A**10 term
    A10 = Float(0.0)
    
    #: A**12 term
    A12 = Float(0.0)
    
    #: A**14 term
    A14 = Float(0.0)
    
    #: A**16 term
    A16 = Float(0.0)
                
    cface = Instance(AsphericFace_, ())
    
    traits_view = View(VGroup(
            Item("z_height", editor=NumEditor, tooltip="surface height at x=0,y=0 in mm"),
            Item("curvature", editor=NumEditor, tooltip="radius of curvature, in mm"),
            Item("conic_const", editor=NumEditor, tooltip="conic constant for the surface"),
            VGrid(
                Item("A4", editor=NumEditor), Item("A6", editor=NumEditor), Item("A8", editor=NumEditor), 
                Item("A10", editor=NumEditor), Item("A12", editor=NumEditor),
                Item("A14", editor=NumEditor), Item("A16", editor=NumEditor),
                show_border=True,
                label="Polynomial coefs"
                ),
            Item("invert", tooltip="Invert the outward sense for this surface")
        ))
    
    @observe("A4, A6, A8, A10, A12, A14, A16")
    def on_coef_change(self, evt):
        setattr(self.cface, evt.name, evt.new)
        self.updated = True
        
    
class DistortionFace(BaseFace):
    """
    This face type wraps an underlying face (the 'base face') and applies a 2D distortion
    function along the local Z-axis.
    """
    #: Z-offset for the face
    z_height = Property()
    
    #: Base face which this object wraps.
    base_face = Instance(BaseFace)
    
    #: The distortion function (e.g. see :py:mod:`rapier.distortions`)
    distortion = Instance(BaseDistortion, ())
    
    cface = Instance(DistortionFace_)
    
    traits_view = View(
                    VGroup(Item("base_face", style="custom"),
                           Item("distortion", style="custom")
                           )
                    )
    
    def _get_z_height(self):
        return self.base_face.z_height
    
    def _cface_default(self):
        return DistortionFace_(base_face=self.base_face.cface, 
                               distortion=self.distortion.c_distortion)
    
    @observe("base_face.updated")
    def on_base_face_changed(self, evt):
        self.cface.base_face = self.base_face.cface
        self.updated = True
        
    @observe("distortion")
    def on_distortion_obj_changed(self, evt):
        self.cface.distortion = self.distortion.c_distortion
        self.updated = True
    
    @observe("distortion.updated")
    def on_distortion_changed(self, evt):
        self.updated = True
    



from traits.api import Float, Str

from .lenses import AsphericLens
from .core.cmaterials import CoatedDispersiveMaterial
from .dispersion import NamedDispersionCurve, NondispersiveCurve


class DispersiveAsphericLens(AsphericLens):
    glass_type = Str("N-BK7")
    coating_thickness = Float(0.283) #microns
    coating_refractive_index = Float(1.37)
    
    def _material_default(self):
        glass = NamedDispersionCurve(self.glass_type)
        cri = self.coating_refractive_index
        coating = NondispersiveCurve(refractive_index=cri)
        air = NondispersiveCurve(1.0)
        
        m = CoatedDispersiveMaterial(dispersion_inside=glass,
                                     dispersion_outside=air,
                                     dispersion_coating=coating)
        return m


class Thorlabs355440(DispersiveAsphericLens):
    name = "Thorlabs#355440"
    diameter = 4.7
    CT = 3.8
    A_curvature = -3.200655
    A_conic = -4.321649
    A4 = -5.521153E-3
    A6 = 1.981378E-3
    A8 = -4.782553E-4
    A10 = 7.328134E-5
    A12 = -5.920460E-6
    A14 = 9.104334E-8
    A16 = 1.291935E-8
    
    B_curvature = 3.200655
    B_conic = -4.321649
    B4 = 5.521153E-3
    B6 = -1.981378E-3
    B8 = 4.782553E-4
    B10 = -7.328134E-5
    B12 = 5.920460E-6
    B14 = -9.104334E-8
    B16 = -1.291935E-8
    glass_type = "D-ZLAF52LA"
    
    
class Thorlabs355440_B(AsphericLens):
    name = "Thorlabs#355440"
    diameter = 4.7
    CT = 3.8
    A_curvature = -3.200655
    A_conic = -4.321649
    A4 = -5.521153E-3
    A6 = 1.981378E-3
    A8 = -4.782553E-4
    A10 = 7.328134E-5
    A12 = -5.920460E-6
    A14 = 9.104334E-8
    A16 = 1.291935E-8
    
    B_curvature = 3.200655
    B_conic = -4.321649
    B4 = 5.521153E-3
    B6 = -1.981378E-3
    B8 = 4.782553E-4
    B10 = -7.328134E-5
    B12 = 5.920460E-6
    B14 = -9.104334E-8
    B16 = -1.291935E-8
    
    coating_thickness = Float(0.283) #microns
    coating_refractive_index = Float(1.37)
    
    def _material_default(self):
        #glass = NondispersiveCurve(1.786) #980nm
        glass = NondispersiveCurve(1.7935) #780nm
        cri = self.coating_refractive_index
        coating = NondispersiveCurve(refractive_index=cri)
        air = NondispersiveCurve(1.0)
        
        m = CoatedDispersiveMaterial(dispersion_inside=glass,
                                     dispersion_outside=air,
                                     dispersion_coating=coating)
        return m
    
    
class Thorlabs352440(AsphericLens):
    """This is an obsolete lens but one we have a lot of experience with which makes it a good test-piece"""
    name = "Thorlabs#352440"
    diameter = 4.7
    CT = 4.06
    A_curvature = -2.40
    A_conic = -4.456395
    A4 = -5.2675614E-03
    A6 = 1.0753165E-03
    A8 = -3.2335319E-04
    A10 = 2.5517886E-05
    B_curvature = 2.39
    B_conic = -0.8371770
    B4 = -7.3100129E-03
    B6 = -3.9100155E-04
    B8 = 1.2597638E-04
    B10 = -2.6451531E-05
    
    coating_thickness = Float(0.283) #microns
    coating_refractive_index = Float(1.37)
    
    ###glass_type = "ECO-550"
    
    def _material_default(self):
        ### ECO-550 is not in the database :(
        glass = NondispersiveCurve(1.594) #Value for 780nm
        cri = self.coating_refractive_index
        coating = NondispersiveCurve(refractive_index=cri)
        air = NondispersiveCurve(1.0)
        
        m = CoatedDispersiveMaterial(dispersion_inside=glass,
                                     dispersion_outside=air,
                                     dispersion_coating=coating)
        return m
    
    
    
    


import os
os.environ['CSF_GraphicShr'] = r"/usr/local/lib/libTKOpenGl.so"

from OCC.Display.wxSamplesGui import display, start_display

from raypier.step_export import make_interp_parabola

para = make_interp_parabola(12.5, 5.0, 15.0, segments=20)

display.DisplayShape(para)

start_display()



from traits.api import Float, on_trait_change, observe

from traitsui.api import View, VGroup, Item

from .bases import Traceable
from .prisms import Extrusion
from .editors import NumEditor, ComplexEditor
from .core.cfaces import ExtrudedPlanarFace
from .core.ctracer import FaceList
from .core.cmaterials import FullDielectricMaterial, \
            LinearPolarisingMaterial, PartiallyReflectiveMaterial, FullDielectricMaterial,\
            SingleLayerCoatedMaterial
            
import numpy



class BaseBeamsplitterCube(Extrusion):
    abstract = True
    size = Float(10.0)
    name = "BaseBeamsplitterCube"
    
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('size', editor=NumEditor),
                       Item('n_inside', editor=ComplexEditor),
                       )
                       )

    @on_trait_change("size")
    def config_profile(self):
        h = self.size/2
        self.z_height_1 = -h
        self.z_height_2 = h
        self.profile = [(-h,-h), (-h,h), (h,h), (h,-h)]


class PolarisingBeamsplitterCube(BaseBeamsplitterCube):
    abstract=False
    name = "Polarising Cube Beamsplitter"
    
    def make_faces(self):
        h = self.size/2
        faces = super(PolarisingBeamsplitterCube,self).make_faces()
        m = LinearPolarisingMaterial()
        p_face = ExtrudedPlanarFace(owner=self, 
                                    z1=self.z_height_1, 
                                    z2=self.z_height_2, 
                                    x1=-h, y1=-h, 
                                    x2=h, y2=h, 
                                    material=m)
        faces.append( p_face )
        return faces
    
    def _vtk_profile(self):
        h = self.size/2
        profile = numpy.concatenate([ self.profile, [(-h,-h), (h,h)] ])
        return profile
    
    
class UnpolarisingBeamsplitterCube(PolarisingBeamsplitterCube):
    abstract=False
    name = "Non_Polarising_Cube_Beamsplitter"
    
    reflectivity = Float(0.5)
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('size', editor=NumEditor),
                       Item('reflectivity', editor=NumEditor),
                       Item('n_inside', editor=ComplexEditor),
                       )
                       )
    
    def _material_default(self):
        #print("n_inside", self.n_inside, "n_outside", self.n_outside)
        m = FullDielectricMaterial(n_inside = self.n_inside,
                                    n_outside = self.n_outside)
#         m = SingleLayerCoatedMaterial(n_inside = self.n_inside,
#                                     n_outside = self.n_outside)
        #print("thresholds:", m.reflection_threshold, m.transmission_threshold)
        return m
    
    def make_faces(self):
        h = self.size/2
        faces = super(PolarisingBeamsplitterCube,self).make_faces()
        m = PartiallyReflectiveMaterial(reflectivity=self.reflectivity)
        p_face = ExtrudedPlanarFace(owner=self, 
                                    z1=self.z_height_1, 
                                    z2=self.z_height_2, 
                                    x1=-h, y1=-h, 
                                    x2=h, y2=h, 
                                    material=m)
        faces.append( p_face )
        return faces
    
    @observe("reflectivity")
    def on_refl_changed(self, evt):
        refl = self.reflectivity
        self.faces.faces[-1].material.reflectivity = refl
        
    


#    Copyright 2009, Teraview Ltd., Bryan Cole
#
#    This file is part of Raypier.
#
#    Raypier is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
a module for parabolic optics e.g. OAPs
"""
from traits.api import HasTraits, Array, Float, Complex,\
            Property, List, Instance, on_trait_change, Range, \
            Tuple, Event, cached_property, Set, Int, Trait, Bool, \
            PrototypedFrom, BaseTuple
from traitsui.api import View, Item, ListEditor, VSplit,\
            RangeEditor, ScrubberEditor, HSplit, VGroup
from tvtk.api import tvtk
import numpy

from raypier.bases import Traceable, normaliseVector, NumEditor,\
            VectorEditor, transformPoints, transformNormals
#from raypier.sources import RayCollection
from raypier.mirrors import BaseMirror
from raypier.core.cfaces import EllipsoidalFace
from raypier.core.ctracer import FaceList

from raypier.vtk_algorithms import EmptyGridSource

class MinMax(BaseTuple):
    def validate(self, obj, name, value):
        valid = super(MinMax,self).validate(obj, name, value)
        a,b = valid
        if a>b:
            valid = (b,a)
        return valid
        


class Ellipsoid(BaseMirror):
    """
    An off-axis ellipsoidal mirror formed into a axis-aligned block.
    """    

    name = "Ellipsoid"
    abstract = False
    
    #: The position of the first focus, in local coordinates.
    focus1 = Tuple(-50.,0.,0.)
    
    #: The position of the second focus, in local coordinates.
    focus2 = Tuple(0., 50., 0.)
    
    #: twice the major axis length, or the distance from one
    #: focus to the ellipsoid edge to the other focus
    size = Float(100.0, desc="twice the major axis length, or the distance from one\
 focus to the ellipsoid edge to the other focus")
    
    #: A 2-tuple giving the lower and upper x-axis bounds of the mirror block. 
    X_bounds = MinMax(-25., 25.)
    
    #: A 2-tuple giving the lower and upper y-axis bounds of the mirror block. 
    Y_bounds = MinMax(-25., 25.)
    
    #: A 2-tuple giving the lower and upper z-axis bounds of the mirror block. 
    Z_bounds = MinMax(0., 50.)
    
    #: Render a glyph at the foci of the ellipsoid.
    show_foci = Bool(True)
    
    foci_Actors = Instance(tvtk.ActorCollection, ())
    
    axes = Property(Tuple, depends_on="focus1, focus2, size",
                 desc="(major, minor) axis lengths")
    
    ellipse_trans = Instance(tvtk.Transform, (), transient=True)
    
    combined_trans = Instance(tvtk.Transform, transient=True)
    
    f1_glyph = Instance(tvtk.SphereSource, (), transient=True)
    f2_glyph = Instance(tvtk.SphereSource, (), transient=True)
    
    f1_act = Instance(tvtk.Follower, (), transient=True)
    f2_act = Instance(tvtk.Follower, (), transient=True)
    
    vtk_grid = Instance(EmptyGridSource, (), transient=True)
    vtk_quadric = Instance(tvtk.Quadric, (), transient=True)
    
    vtkproperty = tvtk.Property(opacity = 0.7,
                             color = (0.8,0.8,0))
    
    traits_view = View(VGroup(
                        Traceable.uigroup,
                        Item("focus1", editor=VectorEditor),
                        Item("focus2", editor=VectorEditor),
                        Item("show_foci"),
                        Item("size", editor=NumEditor),
                        Item("X_bounds"),
                        Item("Y_bounds"),
                        Item("Z_bounds")
                       ),
                       )
    
    def _faces_default(self):
        facelist = FaceList(owner=self)
        facelist.faces=[EllipsoidalFace(owner=self,
                                x1=self.X_bounds[0],
                                x2=self.X_bounds[1],
                                y1=self.Y_bounds[0],
                                y2=self.Y_bounds[1],
                                z1=self.Z_bounds[0],
                                z2=self.Z_bounds[1],
                                major=self.axes[0],
                                minor=self.axes[1])]
        return facelist
    
    @on_trait_change("focus1, focus2, size")
    def config_trans(self, *args):
        f1 = numpy.asarray(self.focus1)
        f2 = numpy.asarray(self.focus2)
        size = self.size
        centre = (f2+f1)/2.
        ellipse_t = self.ellipse_trans
        ellipse_t.identity()
        #ellipse major axis along the X axis
        delta = f2 - f1
        ax = normaliseVector(delta)
        axy = numpy.sqrt(ax[0]**2 + ax[1]**2)
        ellipse_t.rotate_y(numpy.arctan2(ax[2], axy)*180/numpy.pi)
        
        ellipse_t.rotate_z(-numpy.arctan2(ax[1],ax[0])*180/numpy.pi)
        
        ellipse_t.translate(*-centre)
    
    @cached_property
    def _get_axes(self):
        f1 = numpy.asarray(self.focus1)
        f2 = numpy.asarray(self.focus2)
        delta = f2 - f1
        size = self.size
        h = numpy.sqrt((delta**2).sum())
        a = size/2.
        b = 0.5 * numpy.sqrt(size**2 - h**2)
        return (a,b)
    
    def _combined_trans_default(self):
        t = tvtk.Transform()
        t.concatenate(self.transform)
        t.concatenate(self.ellipse_trans.linear_inverse)
        return t
    
    def _show_foci_changed(self, val):
        for act in self.foci_Actors:
            act.visibility = val
        self.render = True
    
    def update_grid(self):
        xmin, xmax = self.X_bounds
        ymin, ymax = self.Y_bounds
        zmin, zmax = self.Z_bounds
        
        source = self.vtk_grid
        size = 20
        dx = (xmax-xmin) / (size-1)
        dy = (ymax-ymin) / (size-1)
        dz = (zmax-zmin) / (size-1)
        source.dimensions = (size,size,size)
        source.origin = (xmin, ymin, zmin)
        source.spacing = (dx, dy, dz)
        
    @on_trait_change("X_bounds, Y_bounds, Z_bounds")
    def change_bounds(self):
        self.update_grid()
        self.vtk_grid.modified()
        self.update=True
    
    @on_trait_change("focus1, focus2, size")
    def config_pipeline(self, *args):
        tp = self.transform.transform_point
        self.f1_act.position = tp(self.focus1)
        self.f2_act.position = tp(self.focus2)
        
        f1 = numpy.asarray(self.focus1)
        f2 = numpy.asarray(self.focus2)
        
        self.f1_glyph.center = f1
        self.f2_glyph.center = f2
        
        centre = (f2+f1)/2.
        ellipse_t = self.ellipse_trans
        ellipse_t.identity()
        #ellipse major axis along the X axis
        delta = f2 - f1
        ax = normaliseVector(delta)
        axy = numpy.sqrt(ax[0]**2 + ax[1]**2)
        ellipse_t.rotate_y(numpy.arctan2(ax[2], axy)*180/numpy.pi)
        
        ellipse_t.rotate_z(-numpy.arctan2(ax[1],ax[0])*180/numpy.pi)
        
        ellipse_t.translate(*-centre)
        
        a,b = self.axes
        #b = 0.5 * numpy.sqrt(size**2 - h**2)  ##not required
        
        q = self.vtk_quadric
        A1 = A2 = 1/(b**2)
        A0 = 1/(a**2)  
        #A8 = 1 
        A9 = -1
        q.coefficients = (A0, A1, A2, 
                          0, 0, 0, 
                          0, 0, 0, 
                          A9)
        self.update=True
        
                                         
    def _pipeline_default(self):
        grid = self.vtk_grid
        #grid.set_execute_method(self.create_grid)
        grid.modified()
        
        quad = self.vtk_quadric
        quad.transform = self.ellipse_trans
        
        clip = tvtk.ClipVolume(input_connection=grid.output_port,
                                 clip_function=quad,
                                 inside_out=0)
        
        topoly = tvtk.GeometryFilter(input_connection=clip.output_port)
        norm = tvtk.PolyDataNormals(input_connection=topoly.output_port)
        transF = tvtk.TransformFilter(input_connection=norm.output_port, transform=self.transform)
        self.config_pipeline()
        grid.modified()
        return transF
    
    def get_actors(self, scene):
        actors = []
        
        sList = [self.f1_glyph, self.f2_glyph]
        cList = [(0,1,0),(1,0,0)]
        
        for s,c in zip(sList, cList):
            s.radius = 1.0
            map = tvtk.PolyDataMapper(input_connection=s.output_port)
            act = tvtk.Actor(mapper=map, user_transform=self.transform)
            act.property.color=c
            actors.append(act)
        
        line = tvtk.LineSource(point1=(-100,0,0),
                               point2=(100,0,0))
        t_line = tvtk.TransformFilter(input_connection=line.output_port,
                                      transform=self.ellipse_trans.linear_inverse)
        map = tvtk.PolyDataMapper(input_connection=t_line.output_port)
        act = tvtk.Actor(mapper=map, user_transform=self.transform)
        act.property.color=(0,0,0)
        actors.append(act)
        
        l1 = tvtk.VectorText(text="F1")
        l2 = tvtk.VectorText(text="F2")
        m1 = tvtk.PolyDataMapper(input_connection=l1.output_port)
        m2 = tvtk.PolyDataMapper(input_connection=l2.output_port)
        act1 = self.f1_act
        act2 = self.f2_act
        
        act1.mapper = m1
        act2.mapper = m2
        
        scale = (5,5,5)
        act1.scale = scale
        act2.scale = scale

        act1.property.color=(0,0,0)
        act2.property.color=(0,0,0)
        
        act1.position = self.focus1
        act2.position = self.focus2
        
        def on_editor(new_ed):
            if new_ed is not None:
                act1.camera = new_ed._camera
                act2.camera = new_ed._camera
        
        scene.on_trait_change(on_editor, "scene_editor")
        
        actors.append(act1)
        actors.append(act2)
        
        for actor in actors:
            self.actors.append(actor)
            self.foci_Actors.append(actor)
            actor.visibility = self.show_foci
        
        return self.actors
    
    def make_step_shape(self):
        from raypier.step_export import make_ellipsoid_mirror
        ell = make_ellipsoid_mirror(self.focus1, 
                                     self.focus2, 
                                     self.size/2.,
                                     self.X_bounds, 
                                     self.Y_bounds, 
                                     self.Z_bounds, 
                                     self.centre, 
                                     self.direction,
                                     self.x_axis)
        return ell, "yellow"



#    Copyright 2009, Teraview Ltd., Bryan Cole
#
#    This file is part of Raypier.
#
#    Raypier is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from traits.api import HasTraits, Event, Str

from raypier.has_queue import HasQueue, on_trait_change

class BaseConstraint(HasQueue):
    update = Event()
    name = Str("a constraint")
    
    
class Constraint(BaseConstraint):
    pass


from traits.api import HasTraits, Instance
from traitsui.api import View

from tvtk.api import tvtk
from tvtk.pyface.scene_model import SceneModel
from tvtk.pyface.scene_editor import SceneEditor


class Tester(HasTraits):
    scene = Instance(SceneModel, (), {'background':(1,1,0.8)}, transient=True)
    
    traits_view = View


from traits.api import Int, Str, HasTraits, Tuple, Array, Callable

from vtk.vtkCommonDataModel import vtkDataObject
from vtk.vtkCommonExecutionModel import vtkAlgorithm
from vtk.vtkCommonExecutionModel import vtkDemandDrivenPipeline
from vtk.vtkCommonExecutionModel import vtkStreamingDemandDrivenPipeline
from vtk.util import numpy_support

import numpy
import vtk
from tvtk.api import tvtk

to_tvtk = tvtk.to_tvtk


class VTKAlgorithm(HasTraits):
    """This is a superclass which can be derived to implement
    Python classes that work with vtkPythonAlgorithm. It implements
    Initialize(), ProcessRequest(), FillInputPortInformation() and
    FillOutputPortInformation().

    Initialize() sets the input and output ports based on data
    members.

    ProcessRequest() calls RequestXXX() methods to implement
    various pipeline passes.

    FillInputPortInformation() and FillOutputPortInformation() set
    the input and output types based on data members.
    """
    
    number_of_input_ports = Int(1)
    input_type = Str("vtkDataSet")
    number_of_output_ports = Int(1)
    output_type = Str("vtkPolyData")
    
#     def __init__(self, **traits):
#         klass = 
#         super(VTKAlgorithm,self).__init__(klass, **traits)

    def Initialize(self, vtkself):
        """Sets up number of input and output ports based on
        NumberOfInputPorts and NumberOfOutputPorts."""

        vtkself.SetNumberOfInputPorts(self.number_of_input_ports)
        vtkself.SetNumberOfOutputPorts(self.number_of_output_ports)

    def GetInputData(self, inInfo, i, j):
        """Convenience method that returns an input data object
        given a vector of information objects and two indices."""

        return inInfo[i].GetInformationObject(j).Get(vtkDataObject.DATA_OBJECT())

    def GetOutputData(self, outInfo, i):
        """Convenience method that returns an output data object
        given an information object and an index."""
        return outInfo.GetInformationObject(i).Get(vtkDataObject.DATA_OBJECT())

    def RequestDataObject(self, vtkself, request, inInfo, outInfo):
        """Overwritten by subclass to manage data object creation.
        There is not need to overwrite this class if the output can
        be created based on the OutputType data member."""
        return 1

    def RequestInformation(self, vtkself, request, inInfo, outInfo):
        """Overwritten by subclass to provide meta-data to downstream
        pipeline."""
        return 1

    def RequestUpdateExtent(self, vtkself, request, inInfo, outInfo):
        """Overwritten by subclass to modify data request going
        to upstream pipeline."""
        return 1

    def RequestData(self, vtkself, request, inInfo, outInfo):
        """Overwritten by subclass to execute the algorithm."""
        raise NotImplementedError('RequestData must be implemented')

    def ProcessRequest(self, vtkself, request, inInfo, outInfo):
        """Splits a request to RequestXXX() methods."""
        print(vtkself)
        print(request)
        print(inInfo)
        print(outInfo)
        if request.Has(vtkDemandDrivenPipeline.REQUEST_DATA_OBJECT()):
            return self.RequestDataObject(vtkself, request, inInfo, outInfo)
        elif request.Has(vtkDemandDrivenPipeline.REQUEST_INFORMATION()):
            return self.RequestInformation(vtkself, request, inInfo, outInfo)
        elif request.Has(vtkStreamingDemandDrivenPipeline.REQUEST_UPDATE_EXTENT()):
            return self.RequestUpdateExtent(vtkself, request, inInfo, outInfo)
        elif request.Has(vtkDemandDrivenPipeline.REQUEST_DATA()):
            return self.RequestData(vtkself, request, inInfo, outInfo)
        return 1

    def FillInputPortInformation(self, vtkself, port, info):
        """Sets the required input type to InputType."""
        info.Set(vtkAlgorithm.INPUT_REQUIRED_DATA_TYPE(), self.input_type)
        return 1

    def FillOutputPortInformation(self, vtkself, port, info):
        """Sets the default output type to OutputType."""
        info.Set(vtkDataObject.DATA_TYPE_NAME(), self.output_type)
        return 1


class PythonAlgorithmBase(tvtk.PythonAlgorithm):
    
    number_of_input_ports = Int(1)
    input_type = Str("vtkDataSet")
    number_of_output_ports = Int(1)
    output_type = Str("vtkPolyData")
    
    class InternalAlgorithm(object):
        "Internal class. Do not use."
        def Initialize(self, vtkself):
            pass

        def FillInputPortInformation(self, vtkself, port, info):
            return to_tvtk(vtkself).FillInputPortInformation(port, info)

        def FillOutputPortInformation(self, vtkself, port, info):
            return to_tvtk(vtkself).FillOutputPortInformation(port, info)

        def ProcessRequest(self, vtkself, request, inInfo, outInfo):
            return to_tvtk(vtkself).ProcessRequest(request, inInfo, outInfo)

    def __init__(self, *args, **kwds):
        """Sets up default NumberOfInputPorts, NumberOfOutputPorts,
        InputType and OutputType that are used by various methods.
        Make sure to call this method from any subclass' __init__"""

        number_of_input_ports = self.number_of_input_ports
        number_of_output_ports = self.number_of_output_ports
        super(PythonAlgorithmBase, self).__init__(*args, **kwds)
        vtk_obj = self._vtk_obj
        vtk_obj.SetPythonObject(PythonAlgorithmBase.InternalAlgorithm())
        self.number_of_input_ports = number_of_input_ports
        self.number_of_output_ports = number_of_output_ports
        

    def get_input_data(self, inInfo, i, j):
        """Convenience method that returns an input data object
        given a vector of information objects and two indices."""

        return inInfo[i].GetInformationObject(j).Get(vtkDataObject.DATA_OBJECT())

    def get_output_data(self, outInfo, i):
        """Convenience method that returns an output data object
        given an information object and an index."""
        return outInfo.GetInformationObject(i).Get(vtkDataObject.DATA_OBJECT())

    def FillInputPortInformation(self, port, info):
        """Sets the required input type to InputType."""
        info.Set(vtkAlgorithm.INPUT_REQUIRED_DATA_TYPE(), self.input_type)
        return 1

    def FillOutputPortInformation(self, port, info):
        """Sets the default output type to OutputType."""
        info.Set(vtkDataObject.DATA_TYPE_NAME(), self.output_type)
        return 1

    def ProcessRequest(self, request, inInfo, outInfo):
        """Splits a request to RequestXXX() methods."""
        if request.Has(vtkDemandDrivenPipeline.REQUEST_DATA_OBJECT()):
            return self.RequestDataObject(request, inInfo, outInfo)
        elif request.Has(vtkDemandDrivenPipeline.REQUEST_INFORMATION()):
            return self.RequestInformation(request, inInfo, outInfo)
        elif request.Has(vtkStreamingDemandDrivenPipeline.REQUEST_UPDATE_EXTENT()):
            return self.RequestUpdateExtent(request, inInfo, outInfo)
        elif request.Has(vtkDemandDrivenPipeline.REQUEST_DATA()):
            return self.RequestData(request, inInfo, outInfo)
        return 1

    def RequestDataObject(self, request, inInfo, outInfo):
        """Overwritten by subclass to manage data object creation.
        There is not need to overwrite this class if the output can
        be created based on the OutputType data member."""
        return 1

    def RequestInformation(self, request, inInfo, outInfo):
        """Overwritten by subclass to provide meta-data to downstream
        pipeline."""
        return 1

    def RequestUpdateExtent(self, request, inInfo, outInfo):
        """Overwritten by subclass to modify data request going
        to upstream pipeline."""
        return 1

    def RequestData(self, request, inInfo, outInfo):
        """Overwritten by subclass to execute the algorithm."""
        raise NotImplementedError('RequestData must be implemented')
        
        
class EmptyGridSource(PythonAlgorithmBase):
    number_of_input_ports = Int(0)
    output_type = Str("vtkStructuredPoints")
    
    dimensions = Tuple(1, 1, 1,)
    origin = Tuple(0.0, 0.0, 0.0)
    spacing = Tuple(0.1, 0.1, 0.1)
    
    def RequestInformation(self, request, inInfo, outInfo):
        dims = self.dimensions
        info = outInfo.GetInformationObject(0)
        info.Set(vtk.vtkStreamingDemandDrivenPipeline.WHOLE_EXTENT(),
            (0, dims[0]-1, 0, dims[1]-1, 0, dims[2]-1), 6)
        return 1
    
    def RequestData(self, request, inInfo, outInfo):
        output = vtk.vtkStructuredPoints.GetData(outInfo)
        dims = self.dimensions
        output.SetDimensions( dims )
        output.SetSpacing( *self.spacing)
        output.SetOrigin( *self.origin)
        return 1
    
    
class NumpyImageSource(EmptyGridSource):
    number_of_input_ports = Int(0)
    output_type = Str("vtkImageData")
    
    dimensions = Tuple(1, 1, 1,)
    origin = Tuple(0.0, 0.0, 0.0)
    spacing = Tuple(0.1, 0.1, 0.1)
    
    image_data = Array #Should be a 2D or 3D array
    
    def _image_data_changed(self, data):
        dims = list(data.shape)
        dims.extend([1,]*(3-len(dims)))
        self.dimensions = tuple(dims)
        self.modified()
        
    def RequestData(self, request, inInfo, outInfo):
        output = vtk.vtkImageData.GetData(outInfo)
        dims = self.dimensions
        output.SetDimensions( dims )
        output.SetSpacing( *self.spacing)
        output.SetOrigin( *self.origin)
        pd = output.GetPointData()
        data = numpy_support.numpy_to_vtk(self.image_data.ravel(), 1)
        pd.SetScalars(data)
        return 1
        
    
if __name__=="__main__":
    from tvtk.pyface.scene_model import SceneModel
    from tvtk.pyface.scene_editor import SceneEditor
    from traits.api import Instance
    from traitsui.api import View, Item
    import vtk
    
    class TestScene(HasTraits):
        scene = Instance(SceneModel, ())
        
        view = View(Item("scene", show_label=False, editor=SceneEditor()))
        
        def show(self):            
            grid = EmptyGridSource(dimensions=(20,30,40))
            src_output_port = grid.output_port
            src_output_port = grid._vtk_obj.GetOutputPort()
            assert src_output_port is not None
            extract = tvtk.ImageDataGeometryFilter(input_connection=src_output_port)
            mapper = tvtk.PolyDataMapper(input_connection=extract.output_port)
            act = tvtk.Actor(mapper=mapper)
            scene = self.scene
            scene.add_actor(act)
            self.configure_traits()
    
    scene = TestScene()
    scene.show()



from raypier.step_export import make_ellipsoid_mirror, export_shapes

f1 = (22.5, 45.1, 50.0)
f2 = (92.5, 45.1, 20.0)

major_axis=52.0

x_bounds = (0.0, 45.0)
y_bounds = (0.0, 45.0)
z_bounds = (0.0, 40.0)

x_axis = (1.0,0.0,0.0)
centre = (0.0,0.0,0.0)
direction = (0.0,0.0,1.0)

e = make_ellipsoid_mirror(f1, f2, major_axis, x_bounds, y_bounds, z_bounds, centre, direction, x_axis)

export_shapes([e], "/home/bryan/EllipsoidMirror.step")





from traits.api import Float, Instance, Enum, Array, on_trait_change

from traitsui.api import View, Item, VGroup, HGroup

from tvtk.api import tvtk
from tvtk.pyface.scene_model import SceneModel
from tvtk.pyface.scene_editor import SceneEditor

from raypier.results import Result
from raypier.fields import EFieldPlane
from raypier.vtk_algorithms import NumpyImageSource
from .core.unwrap2d import unwrap2d

import numpy


class IntensitySurface(Result):
    name = "Field Plane Surface View"
    field_probe = Instance(EFieldPlane)
    
    display = Enum("Intensity", "E_x", "E_y", "E_z", "Phase_x", "Phase_y", "Phase_z")
    
    intensity_data = Array()
    
    scale = Float(1.0)
    
    scene = Instance(SceneModel, transient=True)
    
    #_ds = Instance(tvtk.DepthSortPolyData, transient=True)
    
    _data_source = Instance(NumpyImageSource, (), transient=True)
    _warp = Instance(tvtk.WarpScalar, transient=True)
    _map = Instance(tvtk.PolyDataMapper, transient=True)
    _axes = Instance(tvtk.CubeAxesActor2D, transient=True)
    
    traits_view = View(VGroup(
                        HGroup(
                            Item("display", show_label=False),
                            Item("scale", show_label=False)
                            ),
                        Item("scene", editor=SceneEditor(), show_label=False)
                        ))
    
    def _intensity_data_changed(self, data):
        sdata = data
        self._data_source.image_data = sdata
        map = self._map
        if map is not None:
            map.scalar_range = (sdata.min(), sdata.max())
        self.scene.render()
        
    @on_trait_change("scale, intensity_data")
    def _scale_changed(self, value):
        w = self._warp
        value = self.scale
        low = self.intensity_data.min()
        high = self.intensity_data.max()
        m=self._map
        if m is not None:
            m.scalar_range = (low, high)
        if w is not None:
            w.scale_factor = value/(high-low)
            w.modified()
        
    def _scene_default(self):
        scene = SceneModel(background=(0,0,0))
        
        src = self._data_source
        src.update()
        
        geom = tvtk.ImageDataGeometryFilter(input_connection=src.output_port)
        
        warp = tvtk.WarpScalar(input_connection=geom.output_port,
                               scale_factor=self.scale)
        self._warp = warp
        
        norm = tvtk.PolyDataNormals(input_connection=warp.output_port)
        
#         ###Doesn't work
#         ds = tvtk.DepthSortPolyData(input_connection=norm.output_port,
#                                     camera=tvtk.Camera())
#         self._ds = ds
        
        mapper = tvtk.PolyDataMapper(input_connection=norm.output_port)
        #mapper.scalar_range = (-2.,2.)
        self._map = mapper
        
        self._scale_changed(self.scale)
        
        act = tvtk.Actor(mapper=mapper)
        scene.add_actor(act)
        
        bar = tvtk.ScalarBarActor(lookup_table=mapper.lookup_table,
                                  number_of_labels=5)
        
        scene.add_actor(bar)
        
        axes = tvtk.CubeAxesActor2D(fly_mode="outer_edges",
                                    font_factor=1.5,
                                    scaling=True,
                                    z_axis_visibility=True,
                                    x_label="W",
                                    y_label="H",
                                    z_label="I")
        axes.set_input_connection(norm.output_port)
        self._axes = axes
        #axes.camera = scene.renderer.camera
        scene.add_actor(axes)
        scene.on_trait_change(self.on_activated, "activated")
        
#         def on_sync_camera(camera):
#             print( "CAM")
#             if camera is not None:
#                 ds.camera = camera
#                 
#         scene.on_trait_change(on_sync_camera, "_camera")
        
        return scene
    
    def on_activated(self, obj, val, old, new):
        a = self._axes
        a.camera = obj.camera
        a.use_ranges = True
        
    
    def calc_intensity(self, e_field):
        E = e_field
        mode = self.display
        select = {"x":0, "y":1, "z":2}
        if mode=="Intensity":
            U = (E.real**2).sum(axis=-1) + (E.imag**2).sum(axis=-1)
        elif mode.startswith("E"):
            idx = select[mode[-1]]
            U = E[:,:,idx].real
        elif mode.startswith("Phase"):
            idx = select[mode[-1]]
            e = E[:,:,idx]
            U = numpy.arctan2(e.imag, e.real)
            U, res = unwrap2d(U, anchor=(U.shape[0]//2,U.shape[1]//2))
        self.intensity_data = U
        return U
    
    def _display_changed(self):
        self.on_field_changed(self.field_probe.E_field)
    
    def calc_result(self, model):
        pass
        #self.calc_intensity(self.field_probe.E_field)
    
    def _field_probe_changed(self, old, new):
        if old is not None:
            old.on_trait_change(self.on_field_changed, "E_field", remove=True)
        new.on_trait_change(self.on_field_changed, "E_field")
        
    def on_field_changed(self, e_field):
        self.calc_intensity(e_field)
        p = self.field_probe
        data = self.intensity_data
        self._axes.ranges=[0, p.width*1000, 0, p.height*1000, data.min(), data.max()]
        
            
            
if __name__=="__main__":
    import numpy
    vm = IntensitySurface()
    
    _x = numpy.linspace(-1,1,50)
    _y = numpy.linspace(-1,1,50)
    x,y = numpy.meshgrid(_x,_y)
    
    z = numpy.exp(-(x**2 + y**2)/(0.5**2))
    
    vm.intensity_data = z
    
    vm.configure_traits()
    


'''
Created on 23 Nov 2018

:author: bryan
'''
import pyximport
pyximport.install()

from raypier.core.cmaterials import BaseDispersionCurve

import sqlite3
import pkg_resources
import numpy


GLASS_DATABASE_PATH = "material_data/glass_dispersion_database.db"
MATERIAL_DATABASE = pkg_resources.resource_filename("raypier", GLASS_DATABASE_PATH)


class NondispersiveCurve(BaseDispersionCurve):
    """
    A Dispersion curve for a non-dispersive material with a given refractive index and absorption.
    """
    def __init__(self, refractive_index=1.37, absorption=0.0):
        self._refractive_index=refractive_index
        self._absorption=absorption
        formula_id=0
        coefs = numpy.array([refractive_index,])
        wavelen_min=0.0
        wavelen_max=1000000.0
        super(NondispersiveCurve,self).__init__(formula_id,
                                                  coefs,
                                                  absorption,
                                                  wavelen_min,
                                                  wavelen_max
                                                  )
        
    def __repr__(self):
        return f"<Nondispersion Curve: ri={self._refractive_index}, absorption={self._absorption}>"
        
class FusedSilica(BaseDispersionCurve):
    """
    A Dispersion curve for fused silica.
    """
    def __init__(self, absorption=0.0):
        formula_id=1
        coefs = numpy.array([0.0,
                             0.6961663,
                             0.0684043,
                             0.4079426,
                             0.1162414,
                             0.8974794,
                             9.896161])
        wavelen_min=0.21
        wavelen_max=6.7
        super(FusedSilica,self).__init__(formula_id,
                                                  coefs,
                                                  absorption,
                                                  wavelen_min,
                                                  wavelen_max
                                                  )


class NamedDispersionCurve(BaseDispersionCurve):
    """
    A Dispersion curve obtained from the materials database (http://refractiveindex.info).
    """
    def __init__(self, name=None, book=None, filename=None, absorption=0.0):
        filters = {"name": name, "book": book, "filename": filename}
        names, vals = list(zip(*[('(%s=?)'%(k,),v) for k,v in list(filters.items())\
                             if v is not None]))
        if not names:
            raise ValueError("No material data identifier given. "
                             "Must give at least one name, book or filename")
        where = " AND ".join(names)
        sql = "select * from dispersion where %s"%(where,)
        
        conn = sqlite3.connect(MATERIAL_DATABASE)
        #print(MATERIAL_DATABASE)
        #print("SQL", sql, vals)
        rows = list(conn.execute(sql, vals))
        if len(rows) > 1:
            raise ValueError("Names given return multiple Material entries.")
        
        row=rows[0]
        fname, name, book, formula_id, wavelen_min, wavelen_max, n_coefs = row[:7]
        
        self._data = row
        
        coefs = numpy.array(row[7:7+n_coefs], dtype=numpy.double)
        
        super(NamedDispersionCurve,self).__init__(formula_id,
                                                  coefs,
                                                  absorption,
                                                  wavelen_min,
                                                  wavelen_max
                                                  )
        
    def __repr__(self):
        data = self._data
        return f"<Named Dispersion: name={data[1]}, formula={data[3]}, coefs={data[6]}>"
        
    @classmethod
    def get_glass_names(cls):
        sql  = "select name from dispersion"
        conn = sqlite3.connect(MATERIAL_DATABASE)
        rows = list(conn.execute(sql))
        return [r[0] for r in rows]
        
        
if __name__=="__main__":
    
    BK7 = NamedDispersionCurve("N-LAK22")
    print(BK7.formula_id, list(BK7.coefs))
    print(BK7.evaluate_n([1.0, 1.5]))
    
    print([a for a in NamedDispersionCurve.get_glass_names() if a.startswith("D")])
        


import vtk

source = vtk.vtkProgrammableSource()
def Execute():
    grid = source.GetStructuredGridOutput()
    points = vtk.vtkPoints()

    nx,ny = 25,25
    
    for i in range(nx):
        for j in range(ny):
            points.InsertNextPoint(i/5.0,j/5.0,0.0)
            
    grid.SetDimensions(nx,ny,1)
    grid.SetPoints(points)

source.SetExecuteMethod(Execute)
    
map = vtk.vtkDataSetMapper()
map.SetInputConnection(source.GetOutputPort(2))

act = vtk.vtkActor()
act.SetMapper(map)

ren = vtk.vtkRenderer()
ren.AddActor(act)

renwin = vtk.vtkRenderWindow()
renwin.AddRenderer(ren)

iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renwin)
iren.Start()



import unittest
import numpy

from raypier.sources import ConfocalRayFieldSource
from raypier.mirrors import PlanarWindow
from raypier.tracer import RayTraceModel
from raypier.core.ctracer import cross
from raypier.core.cmaterials import Convert_to_SP


class TestConfocalPhase(unittest.TestCase):
    def test_phase_at_focus(self):
        
        src = ConfocalRayFieldSource(angle=5.0,
                                     angle_step=0.5,
                                     origin=(0,0,0),
                                     working_dist=100.0,
                                     direction=(0,0,1),
                                     wavelength=100.0
                                     )
        rays = src.input_rays.copy_as_array()
        
        vec_to_focus = rays['origin'] - numpy.array([[0,0,src.working_dist]])
        
        path_to_focus = numpy.sqrt((vec_to_focus**2).sum(axis=-1)) 
        offset = 2000.0*path_to_focus*numpy.pi/src.wavelength
        ph_at_origin = rays['phase'] + offset
        
        self.assertTrue( numpy.allclose(ph_at_origin, 0.0))
        
    def test_EField_after_tracing(self):
        src = ConfocalRayFieldSource(angle=5.0,
                                     angle_step=4.0,
                                     origin=(0,0,0),
                                     working_dist=100.0,
                                     direction=(0,0,1),
                                     wavelength=100.0,
                                     )
        
        window = PlanarWindow(centre=(0,0,20),
                              diameter=25.0,
                              thickness=5.0,
                              n_inside=1.0)
        
        model = RayTraceModel(optics=[window],
                              sources=[src])
        model.trace_all()
        
        for rays in src.traced_rays:
            E = rays.E_vector
            H = numpy.cross(rays.direction, E)
            F = E * rays.E1_amp[:,None] + H * rays.E2_amp[:,None]
            print(F.imag)
        
        rays = src.traced_rays
        
        ray1 = rays[0][0]
        ray2 = rays[1][0]
        
        self.assertAlmostEqual(ray1.direction, ray2.direction)
        
        #print(ray2.normals)
        ray2b = Convert_to_SP(ray1, ray2.normals)
        #print(ray1.E_vector, ray2.E_vector)
        
        for ray in (ray1, ray2, ray2b):
            E = ray.E_vector
            H = cross(ray.direction, E)
            F = tuple(e*ray.E1_amp.real + h*ray.E2_amp.real for e,h in zip(E,H))
            print( F)
        
        
         
        
        
        
        



cdef extern from "math.h":
    double M_PI
    double sqrt(double) nogil
    double atan2 (double y, double x ) nogil
    double pow(double x, double y) nogil
    double fabs(double) nogil
    double cos(double) nogil
    double sin(double) nogil
    double acos(double) nogil
    double floor(double) nogil
    int isnan(double) nogil
    int abs(int) nogil

cdef extern from "float.h":
    double DBL_MAX
    
from libc.stdlib cimport malloc, free, realloc

cdef double INF=(DBL_MAX+DBL_MAX)
cdef double NAN=float("NaN")

from .ctracer cimport Face, sep_, \
        vector_t, ray_t, FaceList, subvv_, dotprod_, mag_sq_, norm_,\
            addvv_, multvs_, mag_, transform_t, Transform, transform_c,\
                rotate_c, Shape, Distortion
                
cimport cython
from cython.parallel cimport threadid
cimport numpy as np_
import numpy as np
cimport openmp


num_threads = openmp.omp_get_num_procs()
openmp.omp_set_num_threads(num_threads)
                
                
cdef class SimpleTestZernikeJ7(Distortion):
    """Implement one low-order Zernike poly for testing purposes.
    """
    cdef:
        public double unit_radius
        public double amplitude
        
    def __cinit__(self, **kwds):
        self.unit_radius = kwds.get("unit_radius", 1.0)
        self.amplitude = kwds.get("amplitude", 1.0)
        
    cdef double z_offset_c(self, double x, double y) nogil:
#         cdef:
#             double rho, Z
            
        x /= self.unit_radius
        y /= self.unit_radius
        #rho = sqrt(x*x + y*y)
        #theta = atan2(y, x)
        Z = sqrt(8.0)*(3*(x*x + y*y) - 2)*y
        return Z * self.amplitude
    
    cdef vector_t z_offset_and_gradient_c(self, double x, double y) nogil:
        """The z-axis surface sag is returned as the z-component 
        of the output vector. The x- and y-components of the surface
        gradient are placed in the x- and y- components of vector.
        I.e. the returned vector is
            (dz(x,y)/dx, dz(x,y)/dy, z(x,y))
        """
        cdef:
            double rho, theta, Z, root8=sqrt(8)*self.amplitude, R=self.unit_radius
            vector_t p
            
        x /= R
        y /= R
            
        #rho = sqrt(x*x + y*y)
        #theta = atan2(y, x)
            
        p.x = root8 * 6*x*y /R #root8*(6*y + (3*rho*rho-2)/y)*x
        p.y = root8 * (3*x*x + 9*y*y -2)/R#root8*(6*y*y + 3*rho*rho - 2)
        p.z = root8*(3*(x*x + y*y) - 2)*y 
        return p
    
    
cdef packed struct zernike_coef_t:
    int j
    int n
    int m
    int k
    double value
    
zcoefs_dtype = np.dtype([('j', np.int32),('n',np.int32),('m',np.int32),
                         ('k',np.int32),('value',np.double)])
    
    
cdef struct zer_nmk:
    int n
    int m
    int k
    
    
jnm_map = [(0,0,0)]
    

cdef zer_nmk eval_nmk_c(int j):
    ### Figure out n,m indices from a given j-index
    cdef:
        zer_nmk out
        int n, m, k, next_j, half_n
        
    if j<0:
        raise ValueError("J must be non-negative")
        
    if j >= len(jnm_map):
        n,m,k = jnm_map[-1]
        while True:
            while m<n:
                m += 2
                next_j = (n*(n+2) + m)//2
                half_n = int(floor(n/2.))
                k = half_n*(half_n + 1) + abs(m)
                jnm_map.append( (n,m,k) )
                if next_j == j:
                    out.n = n
                    out.m = m
                    out.k = k
                    return out
                elif next_j > j:
                    raise ValueError("Something went wrong here! Missed j value.")
            n += 1
            m = -2 - n
        
    nmk = jnm_map[j]
    out.n = nmk[0]
    out.m = nmk[1]
    out.k = nmk[2]
    return out


def eval_nmk(int j):
    cdef zer_nmk out
    out = eval_nmk_c(j)
    return (out.n, out.m, out.k)


@cython.boundscheck(False)  # Deactivate bounds checking
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.nonecheck(False)
@cython.cdivision(True)
cdef double zernike_R_c(double r, int k, int n, int m, double[:,:] workspace) nogil:
    cdef:
        int kA, kB, kC, nA, nB, nC, mA, mB, mC , half_n
        double val
        
    if n < m:
        return 0.0
    
    if n == 0:
        return 1.0
    
    val = workspace[0,k]
    if isnan(val):
        nA = n-1
        mA = abs(m-1)
        half_n = nA/2
        kA = half_n*(half_n+1) + abs(mA)#(nA*(nA+2) + mA)//2
        nB = nA
        mB = m+1
        kB = half_n*(half_n+1) + abs(mB) #(nB*(nB+2) + mB)//2
        nC = n-2
        mC = m
        half_n = nC/2
        kC = half_n*(half_n+1) + abs(mC) #(nC*(nC+2) + mC)//2
        val = r*(zernike_R_c(r, kA, nA, mA, workspace) + zernike_R_c(r, kB, nB, mB, workspace))
        val -= zernike_R_c(r, kC, nC, mC, workspace)
        workspace[0,k] = val
        return val
    else:
        return val

    
def zernike_R(double r, int k, int n, int m, double[:,:] workspace):
    return zernike_R_c(r,k,n,m, workspace)
    
    
@cython.boundscheck(False)  # Deactivate bounds checking
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.nonecheck(False)
@cython.cdivision(True)
cdef double zernike_Rprime_c(double r, int k, int n, int m, double[:,:] workspace) nogil:
    cdef:
        int kA, kB, kC, nA, nB, nC, mA, mB, mC, half_n
        double val
        
    if n < m:
        return 0.0
    
    if n == 0:
        return 0.0
    
    val = workspace[1,k]
    if isnan(val):
        nA = n-1
        mA = abs(m-1)
        half_n = nA/2
        kA = half_n*(half_n+1) + abs(mA)
        nB = nA
        mB = m+1
        kB = half_n*(half_n+1) + abs(mB)
        nC = n-2
        mC = m
        kC = (half_n-1)*half_n + abs(mC)
        val = zernike_R_c(r, kA, nA, mA, workspace) + zernike_R_c(r, kB, nB, mB, workspace)
        val += r*( zernike_Rprime_c(r, kA, nA, mA, workspace) + \
                   zernike_Rprime_c(r, kB, nB, mB, workspace) )
        val -= zernike_Rprime_c(r, kC, nC, mC, workspace)
        workspace[1,k] = val
        return val
    else:
        return val
    
    
def zernike_Rprime(double r, int k, int n, int m, double[:,:] workspace):
    return zernike_Rprime_c(r,k,n,m,workspace)


# @cython.boundscheck(False)  # Deactivate bounds checking
# @cython.wraparound(False)   # Deactivate negative indexing.
# @cython.nonecheck(False)
# @cython.cdivision(True)
# cdef double zernike_R_over_r_c(double r, int k, int n, int m, double[:] workspace) nogil:
#     cdef:
#         int kA, kB, kC, nA, nB, nC, mA, mB, mC , half_n
#         double val, val1, val2, t1,t2,t3,t4
#         
#     if n < m:
#         return 0.0
#     
#     if (n == 0) and (r != 0.0):
#         return 1./r
# 
#     nA = n-1
#     mA = abs(m-1)
#     half_n = nA/2
#     kA = half_n*(half_n+1) + abs(mA)#(nA*(nA+2) + mA)//2
#     nB = nA
#     mB = m+1
#     kB = half_n*(half_n+1) + abs(mB) #(nB*(nB+2) + mB)//2
#     nC = n-2
#     mC = m
#     kC = (half_n-1)*half_n + abs(mC) #(nC*(nC+2) + mC)//2
#     
#     t1 = zernike_R_c(r, kA, nA, mA, workspace)
#     t2 = zernike_R_c(r, kB, nB, mB, workspace)
#     val1 = (t1 + t2)
#     
#     t3 = 0.0        
#     if (nA == 0) and (mA == 0):
#         t3 += 1.0
#     elif (nA == 2) and (mA == 0):
#         t3 -= 1.0
#     elif nA>0:
#         t3 += r*zernike_R_over_r_c(r, kA, nA, mA, workspace)
#     val2 = t3
#             
#     t4 = 0.0
#     if (nB == 0) and (mB == 0):
#         t4 += 1.0
#     elif (nB == 2) and (mB == 0):
#         t4 -= 1.0
#     elif nB>0:
#         t4 += r*zernike_R_over_r_c(r, kB, nB, mB, workspace)
#     val2 += t4
#     
#     val = val1
#     
#     val -= zernike_R_over_r_c(r, kC, nC, mC, workspace)
#         
#     with gil:
#         print("nmk:", (n,m,k), (nA,mA,kA),(nB,mB,kB), val1, val2, t1, t2, t3, t4 )
#     
#     return val


@cython.boundscheck(False)  # Deactivate bounds checking
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.nonecheck(False)
@cython.cdivision(True)
cdef double zernike_R_over_r_c(double r, int k, int n, int m, double[:,:] workspace) nogil:
    """
    Results is undefined for m==0.
    """
    cdef:
        int kA, kB, kC, nA, nB, nC, mA, mB, mC , half_n
        double val
        
    if n < m:
        return 0.0
    
    val = workspace[2,k]
    if isnan(val):
        nA = n-1
        mA = abs(m-1)
        half_n = nA/2
        kA = half_n*(half_n+1) + abs(mA)
        nB = nA
        mB = m+1
        kB = half_n*(half_n+1) + abs(mB)
        nC = n-2
        mC = m
        half_n = nC/2
        kC = half_n*(half_n+1) + abs(mC)
        
        val = zernike_R_c(r, kA, nA, mA, workspace) + zernike_R_c(r, kB, nB, mB, workspace) 
        val -= zernike_R_over_r_c(r, kC, nC, mC, workspace)
        workspace[2,k]=val
    return val


def zernike_R_over_r(double r, int k, int n, int m, double[:,:] workspace):
    return zernike_R_over_r_c(r,k,n,m, workspace)

                
cdef class ZernikeDistortion(Distortion):
    cdef:
        public double unit_radius
        
        public int n_coefs, j_max, k_max
        
        public dict coef_map
        zernike_coef_t[:] coefs
        double[:,:,:] workspace
         
    def __cinit__(self, *args, **coefs):
        cdef:
            int n,m,i, j,size, j_max, k_max, half_n, n_max=0
            list clist
            zer_nmk nmk
            double v
            
        self.coef_map = {}
        self.unit_radius = coefs.get("unit_radius", 1.0)
        cdict = {int(k[1:]): float(v) for k,v in coefs.items() if k.startswith("j")}
        if args:
            for k,v in args[0]:
                cdict[int(k)] = float(v)
        self.k_max = 0
        self.set_coefs(list(cdict.items()))
        
    def __getitem__(self, int idx):
        cdef zernike_coef_t out
            
        if idx >= self.n_coefs:
            raise IndexError(f"Index {idx} greater than number of coeffs ({self.n_coefs}).")
        
        out = self.coefs[idx]
        return (out.j, out.n, out.m, out.k, out.value)
        
    def __len__(self):
        return self.n_coefs
    
    property workspace:
        def __get__(self):
            return np.asarray(self.workspace)
        
    def set_coefs(self, coefs):
        cdef:
            int i,j, n_max, k_max
            zernike_coef_t coef
            zer_nmk nmk
            double v
            
        clist = sorted(coefs)
        j_max = max(j for j,v in clist)
        self.j_max = j_max
        self.n_coefs = len(clist)
        self.coefs = np.empty(len(clist), dtype=zcoefs_dtype)
        n_max = 0
        self.coef_map = {}
        for i in range(self.n_coefs):
            j,v = clist[i]
            nmk = eval_nmk_c(j)
            self.coef_map[j]=i
            coef.j = j
            coef.n = nmk.n
            coef.m = nmk.m
            coef.k = nmk.k
            coef.value = v
            self.coefs[i] = coef
            if nmk.n > n_max:
                n_max = nmk.n
        k_max = (n_max//2)*(n_max//2 + 1) + n_max + 1
        if k_max > self.k_max:
            self.workspace = np.empty((num_threads, 3, k_max),'d')
            self.k_max = k_max
    
    def update_coef(self, int j, double value):
        cdef: 
            int i
        try:
            i = self.coef_map[j]
            self.coefs[i].value = value
        except KeyError:
            self.append_coef(j,value)
        
    def get_coef(self, int j):
        cdef int i=self.coef_map[j]
        return self.coefs[i].value
    
    def append_coef(self, int k, double value):
        raise NotImplementedError()
        
    @cython.boundscheck(False)  # Deactivate bounds checking
    @cython.wraparound(False)   # Deactivate negative indexing.
    @cython.nonecheck(False)
    @cython.cdivision(True)
    cdef double z_offset_c(self, double x, double y) nogil:
        cdef:
            double r, theta, Z=0.0, N, PH, R
            int i
            zernike_coef_t coef
            double[:,:] workspace
            
        x /= self.unit_radius
        y /= self.unit_radius
        r = sqrt(x*x + y*y)
        theta = atan2(y, x)
        
        workspace = self.workspace[threadid(),:,:]
        for i in range(self.k_max):
            workspace[0,i] = NAN
        workspace[0,0] = 1.0 #initialise the R_0_0 term
        
        for i in range(self.n_coefs):
            coef = self.coefs[i]
            
            if coef.m==0:
                N = sqrt(coef.n+1)
            else:
                N = sqrt(2*(coef.n+1))
                
            N *= coef.value
                
            if coef.m >= 0:
                PH = cos(coef.m*theta)
            else:
                PH = -sin(coef.m*theta)
            
            R = zernike_R_c(r, coef.k, coef.n, abs(coef.m), workspace)
            Z += N * R * PH 
        return Z
        
    @cython.boundscheck(False)  # Deactivate bounds checking
    @cython.wraparound(False)   # Deactivate negative indexing.
    @cython.nonecheck(False)
    @cython.cdivision(True)
    cdef vector_t z_offset_and_gradient_c(self, double x, double y) nogil:
        """The z-axis surface sag is returned as the z-component 
        of the output vector. The x- and y-components of the surface
        gradient are placed in the x- and y- components of vector.
        I.e. the returned vector is
            (dz(x,y)/dx, dz(x,y)/dy, z(x,y))
        """
        cdef:
            double r, theta, N, PH, R, Rprime, R_over_r
            int i
            int thd = threadid()
            zernike_coef_t coef
            double[:,:] workspace
            vector_t Z
            
        x /= self.unit_radius
        y /= self.unit_radius
        r = sqrt(x*x + y*y)
        theta = atan2(y, x)
        
        workspace = self.workspace[thd,:,:]
        for i in range(self.k_max):
            workspace[0,i] = NAN
            workspace[1,i] = NAN
            workspace[2,i] = NAN
        
        Z.x = 0.0
        Z.y = 0.0
        Z.z = 0.0
        
        for i in range(self.n_coefs):
            coef = self.coefs[i]
                
            if coef.m >= 0:
                PH = cos(coef.m*theta)
                PHprime = -coef.m*sin(coef.m*theta)
            else:
                PH = -sin(coef.m*theta)
                PHprime = -coef.m*cos(coef.m*theta)
            
            R = zernike_R_c(r, coef.k, coef.n, abs(coef.m), workspace)
            Rprime = zernike_Rprime_c(r, coef.k, coef.n, abs(coef.m), workspace)
            R_over_r = zernike_R_over_r_c(r, coef.k, coef.n, abs(coef.m), workspace)
            
            if coef.m==0:
                N = sqrt(coef.n+1)
            else:
                N = sqrt(2*(coef.n+1))
            N *= coef.value
            
            Z.z += N * R * PH
            ### Now eval partial derivatives
            Z.x += N*(Rprime*cos(theta)*PH + R_over_r*(-sin(theta))*PHprime)
            Z.y += N*(Rprime*sin(theta)*PH + R_over_r*(cos(theta))*PHprime)
            
        Z.x /= self.unit_radius
        Z.y /= self.unit_radius
        
        return Z
        